""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.orm import as_selectable
from onegov.core.orm.types import JSON
from onegov.core.upgrade import upgrade_task
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.swissvotes.models import SwissVote
from onegov.swissvotes.models import TranslatablePage
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import select
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import TSVECTOR


@upgrade_task('Rename recommendation columns')
def rename_recommendation_columns(context):
    for old, new in (
        ('ucsp', 'csp'),
        ('sbv', 'sbv_usp'),
        ('cng_travs', 'travs'),
        ('zsa', 'sav')
    ):
        old = f'recommendation_{old}'
        new = f'recommendation_{new}'
        if (
            context.has_column('swissvotes', old)
            and not context.has_column('swissvotes', new)
        ):
            context.operations.alter_column(
                'swissvotes', old, new_column_name=new
            )


@upgrade_task('Add tsvector columns')
def add_tsvector_columns(context):
    for column in ('searchable_text_de_CH', 'searchable_text_fr_CH'):
        if not context.has_column('swissvotes', column):
            context.operations.add_column(
                'swissvotes', Column(column, TSVECTOR())
            )

    for vote in context.app.session().query(SwissVote):
        vote.vectorize_files()


@upgrade_task('Add parliament position columns')
def add_parliament_position_columns(context):
    columns = ('position_national_council', 'position_council_of_states')

    for col in columns:
        if not context.has_column('swissvotes', col):
            context.operations.add_column('swissvotes', Column(col, Integer))

    for vote in context.app.session().query(SwissVote):
        for col in columns:
            yeas = getattr(vote, f'{col}_yeas')
            nays = getattr(vote, f'{col}_nays')
            if yeas is not None and nays is not None:
                if yeas > nays:
                    setattr(vote, f'_{col}', 1)
                elif yeas < nays:
                    setattr(vote, f'_{col}', 2)


@upgrade_task('Add cantonal result columns')
def add_cantonal_result_columns(context):
    for canton in (
        'ag', 'ai', 'ar', 'be', 'bl', 'bs', 'fr', 'ge', 'gl', 'gr', 'ju', 'lu',
        'ne', 'nw', 'ow', 'sg', 'sh', 'so', 'sz', 'tg', 'ti', 'ur', 'vd', 'vs',
        'zg', 'zh'
    ):
        for column in (
            f'result_{canton}_eligible_voters',
            f'result_{canton}_votes_valid',
            f'result_{canton}_votes_total',
            f'result_{canton}_yeas',
            f'result_{canton}_nays',
            f'result_{canton}_accepted'
        ):
            if not context.has_column('swissvotes', column):
                context.operations.add_column(
                    'swissvotes', Column(column, Integer)
                )

        for column in (
            f'result_{canton}_turnout',
            f'result_{canton}_yeas_p'
        ):
            if not context.has_column('swissvotes', column):
                context.operations.add_column(
                    'swissvotes', Column(column, Numeric(13, 10))
                )


@upgrade_task('Add order column to pages')
def add_order_column_to_pagess(context):
    if context.has_column('swissvotes_page', 'order'):
        context.operations.drop_column(
            'swissvotes_page', 'order'
        )
    if not context.has_column('swissvotes_page', 'order'):
        context.operations.add_column(
            'swissvotes_page',
            Column('order', Integer, default=2 ** 16)
        )

        default = (
            'home', 'disclaimer', 'imprint', 'data-protection', 'dataset',
            'about', 'contact'
        )
        pages = TranslatablePageCollection(context.app.session())
        for id in default:
            pages.setdefault(id)
        for order, id in enumerate(default):
            pages.by_id(id).order = order
        for page in pages.query().filter(TranslatablePage.id.notin_(default)):
            order += 1
            page.order = order


@upgrade_task('Add recommendations column')
def add_recommendations_column(context):
    if not context.has_column('swissvotes', 'recommendations'):
        context.operations.add_column(
            'swissvotes',
            Column('recommendations', JSON())
        )

        query = as_selectable("""
            SELECT
                id,  -- Integer
                recommendation_fdp,  -- Integer
                recommendation_cvp,  -- Integer
                recommendation_sps,  -- Integer
                recommendation_svp,  -- Integer
                recommendation_lps,  -- Integer
                recommendation_ldu,  -- Integer
                recommendation_evp,  -- Integer
                recommendation_csp,  -- Integer
                recommendation_pda,  -- Integer
                recommendation_poch,  -- Integer
                recommendation_gps,  -- Integer
                recommendation_sd,  -- Integer
                recommendation_rep,  -- Integer
                recommendation_edu,  -- Integer
                recommendation_fps,  -- Integer
                recommendation_lega,  -- Integer
                recommendation_kvp,  -- Integer
                recommendation_glp,  -- Integer
                recommendation_bdp,  -- Integer
                recommendation_mcg,  -- Integer
                recommendation_sav,  -- Integer
                recommendation_eco,  -- Integer
                recommendation_sgv,  -- Integer
                recommendation_sbv_usp,  -- Integer
                recommendation_sgb,  -- Integer
                recommendation_travs,  -- Integer
                recommendation_vsa  -- Integer
            FROM
                swissvotes
        """)
        session = context.app.session_manager.session()
        for result in session.execute(select(query.c)):
            vote = session.query(SwissVote).filter_by(id=result.id).one()
            vote.recommendations = {
                'fdp': result.recommendation_fdp,
                'cvp': result.recommendation_cvp,
                'sps': result.recommendation_sps,
                'svp': result.recommendation_svp,
                'lps': result.recommendation_lps,
                'ldu': result.recommendation_ldu,
                'evp': result.recommendation_evp,
                'csp': result.recommendation_csp,
                'pda': result.recommendation_pda,
                'poch': result.recommendation_poch,
                'gps': result.recommendation_gps,
                'sd': result.recommendation_sd,
                'rep': result.recommendation_rep,
                'edu': result.recommendation_edu,
                'fps': result.recommendation_fps,
                'lega': result.recommendation_lega,
                'kvp': result.recommendation_kvp,
                'glp': result.recommendation_glp,
                'bdp': result.recommendation_bdp,
                'mcg': result.recommendation_mcg,
                'sav': result.recommendation_sav,
                'eco': result.recommendation_eco,
                'sgv': result.recommendation_sgv,
                'sbv_usp': result.recommendation_sbv_usp,
                'sgb': result.recommendation_sgb,
                'travs': result.recommendation_travs,
                'vsa': result.recommendation_vsa,
            }

    for column in (
        'recommendation_fdp',
        'recommendation_cvp',
        'recommendation_sps',
        'recommendation_svp',
        'recommendation_lps',
        'recommendation_ldu',
        'recommendation_evp',
        'recommendation_csp',
        'recommendation_pda',
        'recommendation_poch',
        'recommendation_gps',
        'recommendation_sd',
        'recommendation_rep',
        'recommendation_edu',
        'recommendation_fps',
        'recommendation_lega',
        'recommendation_kvp',
        'recommendation_glp',
        'recommendation_bdp',
        'recommendation_mcg',
        'recommendation_sav',
        'recommendation_eco',
        'recommendation_sgv',
        'recommendation_sbv_usp',
        'recommendation_sgb',
        'recommendation_travs',
        'recommendation_vsa'
    ):
        if context.has_column('swissvotes', column):
            context.operations.drop_column('swissvotes', column)


@upgrade_task('Add BFS map link columns')
def add_bfs_map_columns(context):
    if not context.has_column('swissvotes', 'bfs_map_de'):
        context.operations.add_column('swissvotes', Column('bfs_map_de', Text))
    if not context.has_column('swissvotes', 'bfs_map_fr'):
        context.operations.add_column('swissvotes', Column('bfs_map_fr', Text))


@upgrade_task('Add other recommendation columns')
def add_other_recommendation_columns(context):
    for name in ('yes', 'no', 'free'):
        column = f'recommendations_other_{name}'
        if not context.has_column('swissvotes', column):
            context.operations.add_column('swissvotes', Column(column, Text))


@upgrade_task('Add divergent recommendations column')
def add_divergent_recommendations_column(context):
    if not context.has_column('swissvotes', 'recommendations_divergent'):
        context.operations.add_column(
            'swissvotes',
            Column('recommendations_divergent', JSON())
        )


@upgrade_task('Rename sbv-usp column')
def rename_sbv_usp_column(context):
    for vote in context.app.session().query(SwissVote):
        if 'sbv_usp' in vote.recommendations:
            vote.recommendations['sbv-usp'] = vote.recommendations['sbv_usp']
            del vote.recommendations['sbv_usp']



from cached_property import cached_property
from collections import OrderedDict
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON
from onegov.file import AssociatedFiles
from onegov.file import File
from onegov.swissvotes import _
from onegov.swissvotes.models.actor import Actor
from onegov.swissvotes.models.canton import Canton
from onegov.swissvotes.models.localized_file import LocalizedFile
from onegov.swissvotes.models.policy_area import PolicyArea
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy_utils import observes
from sqlalchemy.dialects.postgresql import INT4RANGE
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import deferred
from urllib.parse import urlparse
from urllib.parse import urlunparse


class SwissVoteFile(File):
    __mapper_args__ = {'polymorphic_identity': 'swissvote'}


class encoded_property(object):
    """ A shorthand property to return the label of an encoded value. Requires
    the instance the have a `codes`-lookup function. Assumes that the value
    to lookup has the same name as the property prefixed with an underline.

    Example:

        class MyClass(object):
            _value = 0
            value = encoded_property()

            def codes(self, attributes):
                return {0: 'None', 1: 'One'}

    """

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        value = getattr(instance, f'_{self.name}')
        return instance.codes(self.name).get(value)


class SwissVote(Base, TimestampMixin, AssociatedFiles):

    """ A single vote as defined by the code book.

    Some columns are only used when importing/exporting the dataset and are
    lazy loaded.

    """

    __tablename__ = 'swissvotes'

    @staticmethod
    def codes(attribute):
        """ Returns the codes for the given attribute as defined in the code
        book.

        """
        if attribute == 'legal_form':
            return OrderedDict((
                (1, _("Mandatory referendum")),
                (2, _("Optional referendum")),
                (3, _("Popular initiative")),
                (4, _("Direct counter-proposal")),
            ))
        if attribute == 'result_cantons_accepted':
            return OrderedDict((
                (0, _("Rejected")),
                (1, _("Accepted")),
                (3, _("Majority of the cantons not necessary")),
            ))
        if attribute == 'result' or attribute.endswith('_accepted'):
            return OrderedDict((
                (0, _("Rejected")),
                (1, _("Accepted")),
            ))
        if attribute == 'department_in_charge':
            return OrderedDict((
                (1, _("Federal Department of Foreign Affairs (FDFA)")),
                (2, _("Federal Department of Home Affairs (FDHA)")),
                (3, _("Federal Department of Justice and Police (FDJP)")),
                (4, _("Federal Department of Defence, Civil Protection and "
                      "Sport (DDPS)")),
                (5, _("Federal Department of Finance (FDF)")),
                (6, _("Federal Department of Economic Affairs, Education and "
                      "Research (EAER)")),
                (7, _("Federal Department of the Environment, Transport, "
                      "Energy and Communications (DETEC)")),
                (8, _("Federal Chancellery (FCh)")),
            ))
        if attribute == 'position_federal_council':
            return OrderedDict((
                (1, _("Accepting")),
                (2, _("Rejecting")),
                (3, _("Neutral"))
            ))
        if (
            attribute == 'position_parliament'
            or attribute == 'position_national_council'
            or attribute == 'position_council_of_states'
        ):
            return OrderedDict((
                (1, _("Accepting")),
                (2, _("Rejecting")),
            ))
        if attribute == 'recommendation':
            return OrderedDict((
                (1, _("Yea")),
                (2, _("Nay")),
                (3, _("None")),
                (4, _("Empty")),
                (5, _("Free vote")),
                (66, _("Neutral")),
                (9999, _("Organization no longer exists")),
            ))

        raise RuntimeError(f"No codes avaible for '{attribute}'")

    id = Column(Integer, nullable=False, primary_key=True)

    # Formal description
    bfs_number = Column(Numeric(8, 2), nullable=False)
    date = Column(Date, nullable=False)
    decade = Column(INT4RANGE, nullable=False)
    legislation_number = Column(Integer, nullable=False)
    legislation_decade = Column(INT4RANGE, nullable=False)
    title = Column(Text, nullable=False)
    keyword = Column(Text)
    votes_on_same_day = Column(Integer, nullable=False)
    _legal_form = Column('legal_form', Integer, nullable=False)
    legal_form = encoded_property()
    initiator = Column(Text)
    anneepolitique = Column(Text)
    bfs_map_de = Column(Text)
    bfs_map_fr = Column(Text)

    def bfs_map(self, locale):
        return self.bfs_map_fr if locale == 'fr_CH' else self.bfs_map_de

    def bfs_map_host(self, locale):
        try:
            return urlunparse(
                list(urlparse(self.bfs_map(locale))[:2]) + ['', '', '', '']
            )
        except ValueError:
            pass

    # Descriptor
    descriptor_1_level_1 = Column(Numeric(8, 4))
    descriptor_1_level_2 = Column(Numeric(8, 4))
    descriptor_1_level_3 = Column(Numeric(8, 4))
    descriptor_2_level_1 = Column(Numeric(8, 4))
    descriptor_2_level_2 = Column(Numeric(8, 4))
    descriptor_2_level_3 = Column(Numeric(8, 4))
    descriptor_3_level_1 = Column(Numeric(8, 4))
    descriptor_3_level_2 = Column(Numeric(8, 4))
    descriptor_3_level_3 = Column(Numeric(8, 4))

    @cached_property
    def policy_areas(self):
        def get_level(number, level):
            value = getattr(self, f'descriptor_{number}_level_{level}')
            if value is not None:
                return PolicyArea(value, level)

        result = []
        for number in (1, 2, 3):
            for level in (3, 2, 1):
                area = get_level(number, level)
                if area:
                    result.append(area)
                    break
        return result

    # Result
    _result = Column('result', Integer)
    result = encoded_property()
    result_eligible_voters = deferred(Column(Integer), group='dataset')
    result_votes_empty = deferred(Column(Integer), group='dataset')
    result_votes_invalid = deferred(Column(Integer), group='dataset')
    result_votes_valid = deferred(Column(Integer), group='dataset')
    result_votes_total = deferred(Column(Integer), group='dataset')
    result_turnout = Column(Numeric(13, 10))

    _result_people_accepted = Column('result_people_accepted', Integer)
    result_people_accepted = encoded_property()
    result_people_yeas = Column(Integer)
    result_people_nays = Column(Integer)
    result_people_yeas_p = Column(Numeric(13, 10))

    _result_cantons_accepted = Column('result_cantons_accepted', Integer)
    result_cantons_accepted = encoded_property()
    result_cantons_yeas = Column(Numeric(3, 1))
    result_cantons_nays = Column(Numeric(3, 1))
    result_cantons_yeas_p = Column(Numeric(13, 10))

    _result_ag_accepted = Column('result_ag_accepted', Integer)
    result_ag_accepted = encoded_property()
    result_ag_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ag_votes_valid = deferred(Column(Integer), group='dataset')
    result_ag_votes_total = deferred(Column(Integer), group='dataset')
    result_ag_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ag_yeas = deferred(Column(Integer), group='dataset')
    result_ag_nays = deferred(Column(Integer), group='dataset')
    result_ag_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ai_accepted = Column('result_ai_accepted', Integer)
    result_ai_accepted = encoded_property()
    result_ai_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ai_votes_valid = deferred(Column(Integer), group='dataset')
    result_ai_votes_total = deferred(Column(Integer), group='dataset')
    result_ai_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ai_yeas = deferred(Column(Integer), group='dataset')
    result_ai_nays = deferred(Column(Integer), group='dataset')
    result_ai_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ar_accepted = Column('result_ar_accepted', Integer)
    result_ar_accepted = encoded_property()
    result_ar_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ar_votes_valid = deferred(Column(Integer), group='dataset')
    result_ar_votes_total = deferred(Column(Integer), group='dataset')
    result_ar_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ar_yeas = deferred(Column(Integer), group='dataset')
    result_ar_nays = deferred(Column(Integer), group='dataset')
    result_ar_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_be_accepted = Column('result_be_accepted', Integer)
    result_be_accepted = encoded_property()
    result_be_eligible_voters = deferred(Column(Integer), group='dataset')
    result_be_votes_valid = deferred(Column(Integer), group='dataset')
    result_be_votes_total = deferred(Column(Integer), group='dataset')
    result_be_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_be_yeas = deferred(Column(Integer), group='dataset')
    result_be_nays = deferred(Column(Integer), group='dataset')
    result_be_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_bl_accepted = Column('result_bl_accepted', Integer)
    result_bl_accepted = encoded_property()
    result_bl_eligible_voters = deferred(Column(Integer), group='dataset')
    result_bl_votes_valid = deferred(Column(Integer), group='dataset')
    result_bl_votes_total = deferred(Column(Integer), group='dataset')
    result_bl_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_bl_yeas = deferred(Column(Integer), group='dataset')
    result_bl_nays = deferred(Column(Integer), group='dataset')
    result_bl_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_bs_accepted = Column('result_bs_accepted', Integer)
    result_bs_accepted = encoded_property()
    result_bs_eligible_voters = deferred(Column(Integer), group='dataset')
    result_bs_votes_valid = deferred(Column(Integer), group='dataset')
    result_bs_votes_total = deferred(Column(Integer), group='dataset')
    result_bs_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_bs_yeas = deferred(Column(Integer), group='dataset')
    result_bs_nays = deferred(Column(Integer), group='dataset')
    result_bs_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_fr_accepted = Column('result_fr_accepted', Integer)
    result_fr_accepted = encoded_property()
    result_fr_eligible_voters = deferred(Column(Integer), group='dataset')
    result_fr_votes_valid = deferred(Column(Integer), group='dataset')
    result_fr_votes_total = deferred(Column(Integer), group='dataset')
    result_fr_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_fr_yeas = deferred(Column(Integer), group='dataset')
    result_fr_nays = deferred(Column(Integer), group='dataset')
    result_fr_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ge_accepted = Column('result_ge_accepted', Integer)
    result_ge_accepted = encoded_property()
    result_ge_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ge_votes_valid = deferred(Column(Integer), group='dataset')
    result_ge_votes_total = deferred(Column(Integer), group='dataset')
    result_ge_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ge_yeas = deferred(Column(Integer), group='dataset')
    result_ge_nays = deferred(Column(Integer), group='dataset')
    result_ge_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_gl_accepted = Column('result_gl_accepted', Integer)
    result_gl_accepted = encoded_property()
    result_gl_eligible_voters = deferred(Column(Integer), group='dataset')
    result_gl_votes_valid = deferred(Column(Integer), group='dataset')
    result_gl_votes_total = deferred(Column(Integer), group='dataset')
    result_gl_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_gl_yeas = deferred(Column(Integer), group='dataset')
    result_gl_nays = deferred(Column(Integer), group='dataset')
    result_gl_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_gr_accepted = Column('result_gr_accepted', Integer)
    result_gr_accepted = encoded_property()
    result_gr_eligible_voters = deferred(Column(Integer), group='dataset')
    result_gr_votes_valid = deferred(Column(Integer), group='dataset')
    result_gr_votes_total = deferred(Column(Integer), group='dataset')
    result_gr_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_gr_yeas = deferred(Column(Integer), group='dataset')
    result_gr_nays = deferred(Column(Integer), group='dataset')
    result_gr_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ju_accepted = Column('result_ju_accepted', Integer)
    result_ju_accepted = encoded_property()
    result_ju_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ju_votes_valid = deferred(Column(Integer), group='dataset')
    result_ju_votes_total = deferred(Column(Integer), group='dataset')
    result_ju_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ju_yeas = deferred(Column(Integer), group='dataset')
    result_ju_nays = deferred(Column(Integer), group='dataset')
    result_ju_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_lu_accepted = Column('result_lu_accepted', Integer)
    result_lu_accepted = encoded_property()
    result_lu_eligible_voters = deferred(Column(Integer), group='dataset')
    result_lu_votes_valid = deferred(Column(Integer), group='dataset')
    result_lu_votes_total = deferred(Column(Integer), group='dataset')
    result_lu_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_lu_yeas = deferred(Column(Integer), group='dataset')
    result_lu_nays = deferred(Column(Integer), group='dataset')
    result_lu_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ne_accepted = Column('result_ne_accepted', Integer)
    result_ne_accepted = encoded_property()
    result_ne_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ne_votes_valid = deferred(Column(Integer), group='dataset')
    result_ne_votes_total = deferred(Column(Integer), group='dataset')
    result_ne_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ne_yeas = deferred(Column(Integer), group='dataset')
    result_ne_nays = deferred(Column(Integer), group='dataset')
    result_ne_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_nw_accepted = Column('result_nw_accepted', Integer)
    result_nw_accepted = encoded_property()
    result_nw_eligible_voters = deferred(Column(Integer), group='dataset')
    result_nw_votes_valid = deferred(Column(Integer), group='dataset')
    result_nw_votes_total = deferred(Column(Integer), group='dataset')
    result_nw_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_nw_yeas = deferred(Column(Integer), group='dataset')
    result_nw_nays = deferred(Column(Integer), group='dataset')
    result_nw_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ow_accepted = Column('result_ow_accepted', Integer)
    result_ow_accepted = encoded_property()
    result_ow_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ow_votes_valid = deferred(Column(Integer), group='dataset')
    result_ow_votes_total = deferred(Column(Integer), group='dataset')
    result_ow_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ow_yeas = deferred(Column(Integer), group='dataset')
    result_ow_nays = deferred(Column(Integer), group='dataset')
    result_ow_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sg_accepted = Column('result_sg_accepted', Integer)
    result_sg_accepted = encoded_property()
    result_sg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sg_votes_valid = deferred(Column(Integer), group='dataset')
    result_sg_votes_total = deferred(Column(Integer), group='dataset')
    result_sg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sg_yeas = deferred(Column(Integer), group='dataset')
    result_sg_nays = deferred(Column(Integer), group='dataset')
    result_sg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sh_accepted = Column('result_sh_accepted', Integer)
    result_sh_accepted = encoded_property()
    result_sh_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sh_votes_valid = deferred(Column(Integer), group='dataset')
    result_sh_votes_total = deferred(Column(Integer), group='dataset')
    result_sh_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sh_yeas = deferred(Column(Integer), group='dataset')
    result_sh_nays = deferred(Column(Integer), group='dataset')
    result_sh_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_so_accepted = Column('result_so_accepted', Integer)
    result_so_accepted = encoded_property()
    result_so_eligible_voters = deferred(Column(Integer), group='dataset')
    result_so_votes_valid = deferred(Column(Integer), group='dataset')
    result_so_votes_total = deferred(Column(Integer), group='dataset')
    result_so_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_so_yeas = deferred(Column(Integer), group='dataset')
    result_so_nays = deferred(Column(Integer), group='dataset')
    result_so_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_sz_accepted = Column('result_sz_accepted', Integer)
    result_sz_accepted = encoded_property()
    result_sz_eligible_voters = deferred(Column(Integer), group='dataset')
    result_sz_votes_valid = deferred(Column(Integer), group='dataset')
    result_sz_votes_total = deferred(Column(Integer), group='dataset')
    result_sz_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_sz_yeas = deferred(Column(Integer), group='dataset')
    result_sz_nays = deferred(Column(Integer), group='dataset')
    result_sz_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_tg_accepted = Column('result_tg_accepted', Integer)
    result_tg_accepted = encoded_property()
    result_tg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_tg_votes_valid = deferred(Column(Integer), group='dataset')
    result_tg_votes_total = deferred(Column(Integer), group='dataset')
    result_tg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_tg_yeas = deferred(Column(Integer), group='dataset')
    result_tg_nays = deferred(Column(Integer), group='dataset')
    result_tg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ti_accepted = Column('result_ti_accepted', Integer)
    result_ti_accepted = encoded_property()
    result_ti_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ti_votes_valid = deferred(Column(Integer), group='dataset')
    result_ti_votes_total = deferred(Column(Integer), group='dataset')
    result_ti_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ti_yeas = deferred(Column(Integer), group='dataset')
    result_ti_nays = deferred(Column(Integer), group='dataset')
    result_ti_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_ur_accepted = Column('result_ur_accepted', Integer)
    result_ur_accepted = encoded_property()
    result_ur_eligible_voters = deferred(Column(Integer), group='dataset')
    result_ur_votes_valid = deferred(Column(Integer), group='dataset')
    result_ur_votes_total = deferred(Column(Integer), group='dataset')
    result_ur_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_ur_yeas = deferred(Column(Integer), group='dataset')
    result_ur_nays = deferred(Column(Integer), group='dataset')
    result_ur_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_vd_accepted = Column('result_vd_accepted', Integer)
    result_vd_accepted = encoded_property()
    result_vd_eligible_voters = deferred(Column(Integer), group='dataset')
    result_vd_votes_valid = deferred(Column(Integer), group='dataset')
    result_vd_votes_total = deferred(Column(Integer), group='dataset')
    result_vd_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_vd_yeas = deferred(Column(Integer), group='dataset')
    result_vd_nays = deferred(Column(Integer), group='dataset')
    result_vd_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_vs_accepted = Column('result_vs_accepted', Integer)
    result_vs_accepted = encoded_property()
    result_vs_eligible_voters = deferred(Column(Integer), group='dataset')
    result_vs_votes_valid = deferred(Column(Integer), group='dataset')
    result_vs_votes_total = deferred(Column(Integer), group='dataset')
    result_vs_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_vs_yeas = deferred(Column(Integer), group='dataset')
    result_vs_nays = deferred(Column(Integer), group='dataset')
    result_vs_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_zg_accepted = Column('result_zg_accepted', Integer)
    result_zg_accepted = encoded_property()
    result_zg_eligible_voters = deferred(Column(Integer), group='dataset')
    result_zg_votes_valid = deferred(Column(Integer), group='dataset')
    result_zg_votes_total = deferred(Column(Integer), group='dataset')
    result_zg_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_zg_yeas = deferred(Column(Integer), group='dataset')
    result_zg_nays = deferred(Column(Integer), group='dataset')
    result_zg_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    _result_zh_accepted = Column('result_zh_accepted', Integer)
    result_zh_accepted = encoded_property()
    result_zh_eligible_voters = deferred(Column(Integer), group='dataset')
    result_zh_votes_valid = deferred(Column(Integer), group='dataset')
    result_zh_votes_total = deferred(Column(Integer), group='dataset')
    result_zh_turnout = deferred(Column(Numeric(13, 10)), group='dataset')
    result_zh_yeas = deferred(Column(Integer), group='dataset')
    result_zh_nays = deferred(Column(Integer), group='dataset')
    result_zh_yeas_p = deferred(Column(Numeric(13, 10)), group='dataset')

    @cached_property
    def results_cantons(self):
        result = {}
        for canton in Canton.abbreviations():
            value = getattr(self, f'_result_{canton}_accepted')
            if value is not None:
                result.setdefault(value, []).append(Canton(canton))

        codes = self.codes('result_accepted')
        return OrderedDict([
            (codes[key], result[key])
            for key in sorted(result.keys())
        ])

    # Authorities
    _department_in_charge = Column('department_in_charge', Integer)
    department_in_charge = encoded_property()
    procedure_number = Column(Numeric(8, 3))
    _position_federal_council = Column('position_federal_council', Integer)
    position_federal_council = encoded_property()
    _position_parliament = Column('position_parliament', Integer)
    position_parliament = encoded_property()
    _position_national_council = Column('position_national_council', Integer)
    position_national_council = encoded_property()
    position_national_council_yeas = Column(Integer)
    position_national_council_nays = Column(Integer)
    _position_council_of_states = Column('position_council_of_states', Integer)
    position_council_of_states = encoded_property()
    position_council_of_states_yeas = Column(Integer)
    position_council_of_states_nays = Column(Integer)

    # Duration
    duration_federal_assembly = Column(Integer)
    duration_post_federal_assembly = Column(Integer)
    duration_initative_collection = Column(Integer)
    duration_initative_federal_council = Column(Integer)
    duration_initative_total = Column(Integer)
    duration_referendum_collection = Column(Integer)
    duration_referendum_total = Column(Integer)
    signatures_valid = Column(Integer)
    signatures_invalid = Column(Integer)

    # Voting recommendations
    recommendations = Column(JSON, nullable=False, default=dict)
    recommendations_other_yes = Column(Text)
    recommendations_other_no = Column(Text)
    recommendations_other_free = Column(Text)
    recommendations_divergent = Column(JSON, nullable=False, default=dict)

    def get_recommendation(self, name):
        recommendations = self.recommendations or {}
        return self.codes('recommendation').get(recommendations.get(name))

    def group_recommendations(self, recommendations):
        result = {}
        for actor, recommendation in recommendations:
            if recommendation is not None and recommendation != 9999:
                result.setdefault(recommendation, []).append(actor)

        codes = self.codes('recommendation')
        return OrderedDict([
            (codes[key], result[key])
            for key in sorted(result.keys())
        ])

    @cached_property
    def recommendations_parties(self):
        recommendations = self.recommendations or {}
        return self.group_recommendations((
            (Actor(name), recommendations.get(name))
            for name in Actor('').parties
        ))

    @cached_property
    def recommendations_divergent_parties(self):
        recommendations = self.recommendations_divergent or {}
        return self.group_recommendations((
            (
                (Actor(name.split('_')[0]), Canton(name.split('_')[1])),
                recommendation,
            )
            for name, recommendation in sorted(recommendations.items())
        ))

    @cached_property
    def recommendations_associations(self):
        def as_list(value, code):
            return [
                (Actor(name.strip()), code)
                for name in (value or '').split(',')
                if name.strip()
            ]

        recommendations = self.recommendations or {}
        recommendations = [
            (Actor(name), recommendations.get(name))
            for name in Actor('').associations
        ]
        recommendations.extend(as_list(self.recommendations_other_yes, 1))
        recommendations.extend(as_list(self.recommendations_other_no, 2))
        recommendations.extend(as_list(self.recommendations_other_free, 5))
        return self.group_recommendations(recommendations)

    # Electoral strength
    national_council_election_year = Column(Integer)
    national_council_share_fdp = Column(Numeric(13, 10))
    national_council_share_cvp = Column(Numeric(13, 10))
    national_council_share_sp = Column(Numeric(13, 10))
    national_council_share_svp = Column(Numeric(13, 10))
    national_council_share_lps = Column(Numeric(13, 10))
    national_council_share_ldu = Column(Numeric(13, 10))
    national_council_share_evp = Column(Numeric(13, 10))
    national_council_share_csp = Column(Numeric(13, 10))
    national_council_share_pda = Column(Numeric(13, 10))
    national_council_share_poch = Column(Numeric(13, 10))
    national_council_share_gps = Column(Numeric(13, 10))
    national_council_share_sd = Column(Numeric(13, 10))
    national_council_share_rep = Column(Numeric(13, 10))
    national_council_share_edu = Column(Numeric(13, 10))
    national_council_share_fps = Column(Numeric(13, 10))
    national_council_share_lega = Column(Numeric(13, 10))
    national_council_share_kvp = Column(Numeric(13, 10))
    national_council_share_glp = Column(Numeric(13, 10))
    national_council_share_bdp = Column(Numeric(13, 10))
    national_council_share_mcg = Column(Numeric(13, 10))
    national_council_share_ubrige = Column(Numeric(13, 10))
    national_council_share_yeas = Column(Numeric(13, 10))
    national_council_share_nays = Column(Numeric(13, 10))
    national_council_share_neutral = Column(Numeric(13, 10))
    national_council_share_vague = Column(Numeric(13, 10))

    @cached_property
    def has_national_council_share_data(self):
        if self.national_council_election_year:
            return True
        return False

    # attachments
    voting_text = LocalizedFile()
    federal_council_message = LocalizedFile()
    parliamentary_debate = LocalizedFile()
    voting_booklet = LocalizedFile()
    resolution = LocalizedFile()
    realization = LocalizedFile()
    ad_analysis = LocalizedFile()
    results_by_domain = LocalizedFile()

    # searchable attachment texts
    searchable_text_de_CH = Column(TSVECTOR)
    searchable_text_fr_CH = Column(TSVECTOR)

    indexed_files = {
        'voting_text',
        'federal_council_message',
        'parliamentary_debate',
        # we don't include the voting booklet, resolution and ad analysis
        # - they might contain other votes from the same day!
        'realization'
    }

    def vectorize_files(self):
        for locale, language in (('de_CH', 'german'), ('fr_CH', 'french')):
            files = [
                SwissVote.__dict__[file].__get_by_locale__(self, locale)
                for file in self.indexed_files
            ]
            text = ' '.join([f.extract or '' for f in files if f]).strip()
            if text:
                setattr(
                    self,
                    f'searchable_text_{locale}',
                    func.to_tsvector(language, text)
                )

    @observes('files')
    def files_observer(self, files):
        self.vectorize_files()

    def get_file(self, name):
        """ Returns the requested localized file, falls back to the default
        locale.

        """
        fallback = SwissVote.__dict__.get(name).__get_by_locale__(
            self, self.session_manager.default_locale
        )
        return getattr(self, name, None) or fallback

    @property
    def percentages(self):
        """ Returns the positions and results as percentages. """

        def lookup(percentage=None, yeas=None, nays=None, code=None):
            if percentage is not None:
                return round(float(percentage), 1)
            if yeas is not None and nays is not None and (yeas and nays):
                return round(100 * (yeas / (yeas + nays)), 1)
            if code is not None:
                return {0: 0.0, 1: 100.0, 2: 0.0, 3: None}.get(code)
            return None

        result = OrderedDict()
        result[_("Federal Council")] = lookup(
            code=self._position_federal_council
        )
        result[_("National Council")] = lookup(
            yeas=self.position_national_council_yeas,
            nays=self.position_national_council_nays,
            code=self._position_national_council
        )
        result[_("Council of States")] = lookup(
            yeas=self.position_council_of_states_yeas,
            nays=self.position_council_of_states_nays,
            code=self._position_council_of_states
        )
        result[_("Yes Camp")] = lookup(
            percentage=self.national_council_share_yeas
        )
        result[_("People")] = lookup(
            percentage=self.result_people_yeas_p,
            code=self._result_people_accepted
        )
        result[_("Cantons")] = lookup(
            code=self._result_cantons_accepted,
            percentage=self.result_cantons_yeas_p,
            yeas=self.result_cantons_yeas,
            nays=self.result_cantons_nays
        )

        return result



from cgi import FieldStorage
from datetime import date
from decimal import Decimal
from io import BytesIO
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.swissvotes.forms import AttachmentsForm
from onegov.swissvotes.forms import PageForm
from onegov.swissvotes.forms import SearchForm
from onegov.swissvotes.forms import UpdateDatasetForm
from onegov.swissvotes.models import ColumnMapper
from onegov.swissvotes.models import TranslatablePage
from psycopg2.extras import NumericRange
from xlsxwriter.workbook import Workbook


class DummyPrincipal(object):
    pass


class DummyApp(object):
    def __init__(self, session, principal):
        self._session = session
        self.principal = principal

    def session(self):
        return self._session


class DummyRequest(object):
    def __init__(self, session, principal=None, private=False, secret=False):
        self.app = DummyApp(session, principal)
        self.session = session
        self.private = private
        self.secret = secret
        self.locale = 'de_CH'
        self.time_zone = 'Europe/Zurich'

    def is_private(self, model):
        return self.private

    def is_secret(self, model):
        return self.secret

    def include(self, resource):
        pass

    def translate(self, text):
        return text.interpolate()


class DummyPostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


def test_attachments_form(swissvotes_app, attachments):
    session = swissvotes_app.session()
    names = list(attachments.keys())

    # Test apply / update
    votes = SwissVoteCollection(session)
    vote = votes.add(
        id=1,
        bfs_number=Decimal('1'),
        date=date(1990, 6, 2),
        decade=NumericRange(1990, 1999),
        legislation_number=4,
        legislation_decade=NumericRange(1990, 1994),
        title="Vote",
        votes_on_same_day=2,
        _legal_form=1,
    )

    form = AttachmentsForm()
    form.request = DummyRequest(session, DummyPrincipal())

    # ... empty
    form.apply_model(vote)

    assert all([getattr(form, name).data is None for name in names])

    form.update_model(vote)

    assert all([getattr(vote, name) is None for name in names])

    # ... add attachments (de_CH)
    for name, attachment in attachments.items():
        setattr(vote, name, attachment)
        session.flush()

    # ... not present with fr_CH
    vote.session_manager.current_locale = 'fr_CH'

    form.apply_model(vote)

    assert all([getattr(form, name).data is None for name in names])

    # ... present with de_CH
    vote.session_manager.current_locale = 'de_CH'

    form.apply_model(vote)

    for name in names:
        data = getattr(form, name).data
        assert data['data']
        assert data['size']
        assert data['filename'] == name
        assert data['mimetype'] in (
            'application/pdf',
            'application/zip',
            'application/vnd.ms-office',
            'application/octet-stream',
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        )

    form.update_model(vote)

    for name in names:
        file = getattr(vote, name)
        assert file == attachments[name]
        assert file.reference.filename == name
        assert file.reference.content_type in (
            'application/pdf',
            'application/zip',
            'application/vnd.ms-office',
            'application/octet-stream',
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        )

    # ... replace all de_CH
    for name in names:
        field_storage = FieldStorage()
        field_storage.file = BytesIO(f'{name}-1'.encode())
        field_storage.type = 'image/png'  # ignored
        field_storage.filename = f'{name}.png'  # extension removed

        getattr(form, name).process(DummyPostData({name: field_storage}))

    form.update_model(vote)

    assert all([
        getattr(vote, name).reference.file.read() == f'{name}-1'.encode()
        for name in names
    ])

    # ... add all fr_CH
    vote.session_manager.current_locale = 'fr_CH'

    for name in names:
        field_storage = FieldStorage()
        field_storage.file = BytesIO(f'{name}-fr'.encode())
        field_storage.type = 'image/png'  # ignored
        field_storage.filename = f'{name}.png'  # extension removed

        getattr(form, name).process(DummyPostData({name: field_storage}))

    form.update_model(vote)

    assert all([
        getattr(vote, name).reference.file.read() == f'{name}-fr'.encode()
        for name in names
    ])

    # ... delete all fr_CH
    for name in names:
        getattr(form, name).action = 'delete'

    form.update_model(vote)

    assert all([getattr(vote, name) is None for name in names])

    vote.session_manager.current_locale = 'de_CH'

    assert all([getattr(vote, name) is not None for name in names])

    # ... delete all de_CH
    for name in names:
        getattr(form, name).action = 'delete'

    form.update_model(vote)

    assert vote.files == []

    # Test validation
    form = AttachmentsForm()
    assert form.validate()


def test_page_form(session):
    # Test apply / update
    pages = TranslatablePageCollection(session)
    page = pages.add(
        id='page',
        title_translations={'de_CH': 'Titel', 'en_US': 'Title'},
        content_translations={'de_CH': 'Inhalt', 'en_US': 'Content'},
    )

    form = PageForm()

    # .. de_CH
    form.apply_model(page)
    assert form.title.data == 'Titel'
    assert form.content.data == 'Inhalt'

    form.title.data = 'A'
    form.content.data = 'B'

    form.update_model(page)

    assert page.title_translations == {'de_CH': 'A', 'en_US': 'Title'}
    assert page.content_translations == {'de_CH': 'B', 'en_US': 'Content'}
    assert page.id == 'page'

    # ... en_US
    page.session_manager.current_locale = 'en_US'

    form.apply_model(page)
    assert form.title.data == 'Title'
    assert form.content.data == 'Content'

    form.title.data = 'C'
    form.content.data = 'D'

    form.update_model(page)

    assert page.title_translations == {'de_CH': 'A', 'en_US': 'C'}
    assert page.content_translations == {'de_CH': 'B', 'en_US': 'D'}
    assert page.id == 'page'

    # ... fr_CH
    page.session_manager.current_locale = 'fr_CH'

    form.apply_model(page)
    assert form.title.data == 'A'
    assert form.content.data == 'B'

    form.title.data = 'E'
    form.content.data = 'F'

    form.update_model(page)

    assert page.title_translations == {
        'de_CH': 'A', 'en_US': 'C', 'fr_CH': 'E'
    }
    assert page.content_translations == {
        'de_CH': 'B', 'en_US': 'D', 'fr_CH': 'F'
    }
    assert page.id == 'page'

    # Test ID generation
    form = PageForm()
    form.request = DummyRequest(session, DummyPrincipal())

    assert form.id == 'page-1'

    form.title.data = ' Über uns '
    assert form.id == 'uber-uns'

    form.update_model(page)
    assert page.id == 'page'

    page = TranslatablePage()
    form.update_model(page)
    assert page.id == 'uber-uns'

    # Test validation
    form = PageForm()
    assert not form.validate()

    form = PageForm(DummyPostData({'title': 'X', 'content': 'Y'}))
    assert form.validate()


def test_search_form(session):
    # Test on request / popluate
    form = SearchForm()
    form.request = DummyRequest(session, DummyPrincipal())
    form.on_request()

    assert form.legal_form.choices == [
        (1, 'Mandatory referendum'),
        (2, 'Optional referendum'),
        (3, 'Popular initiative'),
        (4, 'Direct counter-proposal')
    ]
    assert form.result.choices == [
        (0, 'Rejected'),
        (1, 'Accepted')
    ]
    assert form.position_federal_council.choices == [
        (1, 'Accepting'),
        (2, 'Rejecting'),
        (3, 'Neutral'),
        (-1, 'Without / Unknown')
    ]
    assert form.position_national_council.choices == [
        (1, 'Accepting'),
        (2, 'Rejecting'),
        (-1, 'Without / Unknown')
    ]
    assert form.position_council_of_states.choices == [
        (1, 'Accepting'),
        (2, 'Rejecting'),
        (-1, 'Without / Unknown')
    ]
    assert form.policy_area.choices == []

    # ... descriptors available
    votes = SwissVoteCollection(session)
    kwargs = {
        'date': date(1990, 6, 2),
        'decade': NumericRange(1990, 1999),
        'legislation_number': 4,
        'legislation_decade': NumericRange(1990, 1994),
        'title': "Vote",
        'votes_on_same_day': 2,
        '_legal_form': 1,
    }
    votes.add(
        id=1,
        bfs_number=Decimal('1'),
        descriptor_1_level_1=Decimal('4'),
        descriptor_1_level_2=Decimal('4.2'),
        descriptor_1_level_3=Decimal('4.21'),
        descriptor_2_level_1=Decimal('10'),
        descriptor_2_level_2=Decimal('10.3'),
        descriptor_2_level_3=Decimal('10.35'),
        descriptor_3_level_1=Decimal('10'),
        descriptor_3_level_2=Decimal('10.3'),
        descriptor_3_level_3=Decimal('10.33'),
        **kwargs
    )
    votes.add(
        id=2,
        bfs_number=Decimal('3'),
        descriptor_1_level_1=Decimal('8'),
        descriptor_2_level_1=Decimal('7'),
        descriptor_2_level_2=Decimal('7.5'),
        descriptor_3_level_1=Decimal('9'),
        descriptor_3_level_2=Decimal('9.3'),
        descriptor_3_level_3=Decimal('9.39'),
        **kwargs
    )
    votes.add(
        id=3,
        bfs_number=Decimal('3'),
        descriptor_1_level_1=Decimal('8'),
        descriptor_1_level_2=Decimal('8.7'),
        descriptor_2_level_1=Decimal('6'),
        descriptor_2_level_2=Decimal('6.1'),
        descriptor_2_level_3=Decimal('6.14'),
        **kwargs
    )

    form.on_request()
    assert form.policy_area.choices == [
        ('4', 'd-1-4'),
        ('4.42', 'd-2-42'),
        ('4.42.421', 'd-3-421'),
        ('6', 'd-1-6'),
        ('6.61', 'd-2-61'),
        ('6.61.614', 'd-3-614'),
        ('7', 'd-1-7'),
        ('7.75', 'd-2-75'),
        ('8', 'd-1-8'),
        ('8.87', 'd-2-87'),
        ('9', 'd-1-9'),
        ('9.93', 'd-2-93'),
        ('10', 'd-1-10'),
        ('10.103', 'd-2-103'),
        ('10.103.1033', 'd-3-1033'),
        ('10.103.1035', 'd-3-1035')
    ]

    # Test apply
    form.apply_model(votes)
    assert form.from_date.data is None
    assert form.to_date.data is None
    assert form.legal_form.data == [1, 2, 3, 4]
    assert form.result.data == [0, 1]
    assert form.policy_area.data is None
    assert form.term.data is None
    assert form.full_text.data == 1
    assert form.position_federal_council.data == [1, 2, 3, -1]
    assert form.position_national_council.data == [1, 2, -1]
    assert form.position_council_of_states.data == [1, 2, - 1]
    assert form.sort_by.data is None
    assert form.sort_order.data is None

    votes.from_date = date(2010, 1, 1)
    votes.to_date = date(2010, 12, 31)
    votes.legal_form = [1, 2]
    votes.result = [0]
    votes.policy_area = ['6', '7.75', '10.103.1035']
    votes.term = 'term'
    votes.full_text = False
    votes.position_federal_council = [2, 3]
    votes.position_national_council = [-1]
    votes.position_council_of_states = [-1, 1, 2]
    votes.sort_by = 'title'
    votes.sort_order = 'ascending'

    form.apply_model(votes)

    assert form.from_date.data == date(2010, 1, 1)
    assert form.to_date.data == date(2010, 12, 31)
    assert form.legal_form.data == [1, 2]
    assert form.result.data == [0]
    assert form.policy_area.data == ['6', '7.75', '10.103.1035']
    assert form.term.data == 'term'
    assert form.full_text.data == 0
    assert form.position_federal_council.data == [2, 3]
    assert form.position_national_council.data == [-1]
    assert form.position_council_of_states.data == [-1, 1, 2]
    assert form.sort_by.data == 'title'
    assert form.sort_order.data == 'ascending'

    # Test validation
    form = SearchForm()
    assert form.validate()


def test_update_dataset_form(session):
    request = DummyRequest(session, DummyPrincipal())

    # Validate
    form = UpdateDatasetForm()
    form.request = request
    assert not form.validate()

    file = BytesIO()
    workbook = Workbook(file)
    worksheet = workbook.add_worksheet()
    worksheet.write_row(0, 0, ColumnMapper().columns.values())
    worksheet.write_row(1, 0, [
        100.1,  # anr / NUMERIC
        '1.2.2008',  # datum / DATE
        1,  # legislatur / INTEGER
        '2004-2008',  # legisjahr / INT4RANGE
        '2000-2009',  # jahrzent / INT4RANGE
        'titel',  # titel / TEXT
        'stichwort',  # stichwort / TEXT
        2,  # anzahl / INTEGER
        3,  # rechtsform
    ])
    workbook.close()
    file.seek(0)

    field_storage = FieldStorage()
    field_storage.file = file
    field_storage.type = 'application/excel'
    field_storage.filename = 'test.xlsx'

    form.dataset.process(DummyPostData({'dataset': field_storage}))

    assert form.validate()



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.swissvotes import _
from onegov.swissvotes.layouts.default import DefaultLayout


class VotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Votes")

    @cached_property
    def editbar_links(self):
        result = []
        if self.request.has_role('admin', 'editor'):
            result.append(
                Link(
                    text=_("Update dataset"),
                    url=self.request.link(self.model.default(), name='update'),
                    attrs={'class': 'upload-icon'}
                )
            )
            result.append(
                Link(
                    text=_("Download dataset (CSV)"),
                    url=self.request.link(self.model.default(), name='csv'),
                    attrs={'class': 'export-icon'}
                )
            )
            result.append(
                Link(
                    text=_("Download dataset (XLSX)"),
                    url=self.request.link(self.model.default(), name='xlsx'),
                    attrs={'class': 'export-icon'}
                )
            )
        if self.request.has_role('admin'):
            result.append(
                Link(
                    text=_("Delete all votes"),
                    url=self.request.link(self.model.default(), name='delete'),
                    attrs={'class': 'delete-icon'}
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.votes_url),
        ]


class UpdateVotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Update dataset")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.title, '#'),
        ]


class DeleteVotesLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Delete all votes")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Votes"), self.votes_url),
            Link(self.title, '#'),
        ]



from datetime import date
from decimal import Decimal
from onegov.swissvotes.models import SwissVote
from psycopg2.extras import NumericRange
from transaction import commit
from webtest import TestApp as Client
from webtest.forms import Upload


def test_view_vote(swissvotes_app):
    swissvotes_app.session().add(
        SwissVote(
            bfs_number=Decimal('100.1'),
            date=date(1990, 6, 2),
            decade=NumericRange(1990, 1999),
            legislation_number=4,
            legislation_decade=NumericRange(1990, 1994),
            title="Vote",
            keyword="Keyword",
            votes_on_same_day=2,
            _legal_form=1,
            initiator="Initiator",
            anneepolitique="anneepolitique",
            descriptor_1_level_1=Decimal('4'),
            descriptor_1_level_2=Decimal('4.2'),
            descriptor_1_level_3=Decimal('4.21'),
            descriptor_2_level_1=Decimal('10'),
            descriptor_2_level_2=Decimal('10.3'),
            descriptor_2_level_3=Decimal('10.35'),
            descriptor_3_level_1=Decimal('10'),
            descriptor_3_level_2=Decimal('10.3'),
            descriptor_3_level_3=Decimal('10.33'),
            _result=1,
            result_eligible_voters=2,
            result_votes_empty=3,
            result_votes_invalid=4,
            result_votes_valid=5,
            result_votes_total=6,
            result_turnout=Decimal('20.01'),
            _result_people_accepted=1,
            result_people_yeas=8,
            result_people_nays=9,
            result_people_yeas_p=Decimal('40.01'),
            _result_cantons_accepted=1,
            result_cantons_yeas=Decimal('1.5'),
            result_cantons_nays=Decimal('24.5'),
            result_cantons_yeas_p=Decimal('60.01'),
            _department_in_charge=1,
            procedure_number=Decimal('24.557'),
            _position_federal_council=1,
            _position_parliament=1,
            _position_national_council=1,
            position_national_council_yeas=10,
            position_national_council_nays=20,
            _position_council_of_states=1,
            position_council_of_states_yeas=30,
            position_council_of_states_nays=40,
            duration_federal_assembly=30,
            duration_post_federal_assembly=31,
            duration_initative_collection=32,
            duration_initative_federal_council=33,
            duration_initative_total=34,
            duration_referendum_collection=35,
            duration_referendum_total=36,
            signatures_valid=40,
            signatures_invalid=41,
            recommendations={
                'fdp': 1,
                'cvp': 1,
                'sps': 1,
                'svp': 1,
                'lps': 2,
                'ldu': 2,
                'evp': 2,
                'csp': 3,
                'pda': 3,
                'poch': 3,
                'gps': 4,
                'sd': 4,
                'rep': 4,
                'edu': 5,
                'fps': 5,
                'lega': 5,
                'kvp': 66,
                'glp': 66,
                'bdp': None,
                'mcg': 9999,
                'sav': 1,
                'eco': 2,
                'sgv': 3,
                'sbv-usp': 3,
                'sgb': 3,
                'travs': 3,
                'vsa': 9999,
            },
            recommendations_other_yes="Pro Velo",
            recommendations_other_no=None,
            recommendations_other_free="Pro Natura, Greenpeace",
            recommendations_divergent={
                'fdp-fr_ch': 2,
                'jcvp_ch': 2,
            },
            national_council_election_year=1990,
            national_council_share_fdp=Decimal('01.10'),
            national_council_share_cvp=Decimal('02.10'),
            national_council_share_sp=Decimal('03.10'),
            national_council_share_svp=Decimal('04.10'),
            national_council_share_lps=Decimal('05.10'),
            national_council_share_ldu=Decimal('06.10'),
            national_council_share_evp=Decimal('07.10'),
            national_council_share_csp=Decimal('08.10'),
            national_council_share_pda=Decimal('09.10'),
            national_council_share_poch=Decimal('10.10'),
            national_council_share_gps=Decimal('11.10'),
            national_council_share_sd=Decimal('12.10'),
            national_council_share_rep=Decimal('13.10'),
            national_council_share_edu=Decimal('14.10'),
            national_council_share_fps=Decimal('15.10'),
            national_council_share_lega=Decimal('16.10'),
            national_council_share_kvp=Decimal('17.10'),
            national_council_share_glp=Decimal('18.10'),
            national_council_share_bdp=Decimal('19.10'),
            national_council_share_mcg=Decimal('20.20'),
            national_council_share_ubrige=Decimal('21.20'),
            national_council_share_yeas=Decimal('22.20'),
            national_council_share_nays=Decimal('23.20'),
            national_council_share_neutral=Decimal('24.20'),
            national_council_share_vague=Decimal('25.10'),
        )
    )
    commit()

    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    page = client.get('/').maybe_follow().click("Abstimmungen")
    page = page.click("Details")
    assert "100.1" in page
    assert "Vote" in page
    assert "Keyword" in page
    assert "02.06.1990" in page
    assert "Obligatorisches Referendum" in page
    assert (
        "Wirtschaft &gt; Arbeit und Beschäftigung &gt; Arbeitsbedingungen"
    ) in page
    assert (
        "Soziale Fragen – Sozialpolitik &gt; Soziale Gruppen &gt; "
        "Kinder und Jugendliche"
    ) in page
    assert (
        "Soziale Fragen – Sozialpolitik &gt; Soziale Gruppen &gt; "
        "Stellung der Frau"
    ) in page
    assert "anneepolitique" in page
    assert "Befürwortend" in page
    assert "24.557" in page
    assert "30 Tage" in page
    assert "(10 Ja, 20 Nein)" in page
    assert "(30 Ja, 40 Nein)" in page
    assert "CVP" in page
    assert "FDP" in page
    assert "SPS" in page
    assert "SVP" in page
    assert "EVP" in page
    assert "LdU" in page
    assert "LPS" in page
    assert "CSP" in page
    assert "PdA" in page
    assert "POCH" in page
    assert "GPS" in page
    assert "REP" in page
    assert "SD" in page
    assert "EDU" in page
    assert "FPS" in page
    assert "Lega" in page
    assert "GLP" in page
    assert "KVP" in page
    assert "SAV" in page
    assert "eco" in page
    assert "SBV" in page
    assert "SGB" in page
    assert "SGV" in page
    assert "TravS" in page
    assert "Pro Velo" in page
    assert "Pro Natura" in page
    assert "Greenpeace" in page
    assert "FDP.Die Liberalen Frauen" in page
    assert "Junge CVP" in page
    assert "22.2%" in page
    assert "Details" in page
    assert "Angenommen" in page
    assert "(40.01% Jastimmen)" in page
    assert "(1.5 Ja, 24.5 Nein)" in page
    assert "20.01%" in page

    swissvotes_app.session().query(SwissVote).one()._legal_form = 3
    commit()

    page = client.get('/').maybe_follow().click("Abstimmungen")
    page = page.click("Details")
    assert "Volksinitiative" in page
    assert "Initiator" in page
    assert "32 Tage" in page

    # Party strengths
    page = page.click("Details")
    assert "Nationalratswahl 1990"
    assert "21.2%" in page
    assert "22.2%" in page
    assert "23.2%" in page
    assert "24.2%" in page
    assert "25.1%" in page
    assert "1.1%" in page
    assert "2.1%" in page
    assert "3.1%" in page
    assert "4.1%" in page
    assert "5.1%" in page
    assert "6.1%" in page
    assert "7.1%" in page
    assert "8.1%" in page
    assert "9.1%" in page
    assert "10.1%" in page
    assert "11.1%" in page
    assert "12.1%" in page
    assert "13.1%" in page
    assert "14.1%" in page
    assert "15.1%" in page
    assert "16.1%" in page
    assert "17.1%" in page
    assert "18.1%" in page
    assert "19.1%" in page
    assert "20.2%" in page

    # Percentages
    page = client.get(page.request.url.replace('/strengths', '/percentages'))
    assert page.json == {
        'results': [
            {'text': 'Bundesrat', 'value': 100.0},
            {'text': 'Nationalrat', 'value': 33.3},
            {'text': 'Ständerat', 'value': 42.9},
            {'text': 'Ja-Lager', 'value': 22.2},
            {'text': 'Volk', 'value': 40.0},
            {'text': 'Stände', 'value': 60.0}
        ],
        'title': 'Vote'
    }

    # Delete vote
    login = client.get('/auth/login')
    login.form['username'] = 'admin@example.org'
    login.form['password'] = 'hunter2'
    login.form.submit()

    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Details").click("Abstimmung löschen")
    manage = manage.form.submit().follow()

    assert swissvotes_app.session().query(SwissVote).count() == 0


def test_vote_upload(swissvotes_app, attachments):
    names = attachments.keys()

    swissvotes_app.session().add(
        SwissVote(
            bfs_number=Decimal('100.1'),
            date=date(1990, 6, 2),
            decade=NumericRange(1990, 1999),
            legislation_number=4,
            legislation_decade=NumericRange(1990, 1994),
            title="Vote",
            keyword="Keyword",
            votes_on_same_day=2,
            _legal_form=3,
            initiator="Initiator",
        )
    )
    commit()

    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    login = client.get('/auth/login')
    login.form['username'] = 'admin@example.org'
    login.form['password'] = 'hunter2'
    login.form.submit()

    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Details").click("Anhänge verwalten")
    for name in names:
        manage.form[name] = Upload(
            f'{name}.png',  # ignored
            attachments[name].reference.file.read(),
            'image/png'  # ignored
        )
    manage = manage.form.submit().follow()
    assert "Anhänge aktualisiert" in manage

    for name in names:
        name = name.replace('_', '-')
        page = client.get(manage.pyquery(f'a.{name}')[0].attrib['href'])
        assert page.content_type in (
            'application/pdf',
            'application/zip',
            'application/vnd.ms-office',
            'application/octet-stream',
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        )
        assert page.content_length
        assert page.body
        assert page.content_disposition.startswith('inline; filename=100.1')

    # Fallback
    client.get('/locale/en_US').follow()
    manage = client.get('/').maybe_follow().click("Votes")
    manage = manage.click("Details")
    for name in names:
        name = name.replace('_', '-')
        page = client.get(manage.pyquery(f'a.{name}')[0].attrib['href'])
        assert page.content_type in (
            'application/pdf',
            'application/zip',
            'application/vnd.ms-office',
            'application/octet-stream',
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        )
        assert page.content_length
        assert page.body
        assert page.content_disposition.startswith('inline; filename=100.1')


def test_vote_pagination(swissvotes_app):
    for day, number in ((1, '100'), (2, '101.1'), (2, '101.2'), (3, '102')):
        swissvotes_app.session().add(
            SwissVote(
                bfs_number=Decimal(number),
                date=date(1990, 6, day),
                decade=NumericRange(1990, 1999),
                legislation_number=4,
                legislation_decade=NumericRange(1990, 1994),
                title="Vote",
                keyword="Keyword",
                votes_on_same_day=2,
                _legal_form=3,
                initiator="Initiator",
            )
        )
    commit()

    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    # 102
    page = client.get('/').maybe_follow().click("Abstimmungen")
    page = page.click("Details", index=0)
    assert "<td>102</td>" in page
    assert "Vorherige Abstimmung" in page
    assert "Nächste Abstimmung" not in page

    # 101.2
    page = page.click("Vorherige Abstimmung")
    assert "<td>101.2</td>" in page
    assert "Vorherige Abstimmung" in page
    assert "Nächste Abstimmung" in page

    # 101.1
    page = page.click("Vorherige Abstimmung")
    assert "<td>101.1</td>" in page
    assert "Vorherige Abstimmung" in page
    assert "Nächste Abstimmung" in page

    # 100
    page = page.click("Vorherige Abstimmung")
    assert "<td>100</td>" in page
    assert "Vorherige Abstimmung" not in page
    assert "Nächste Abstimmung" in page

    # 101.1
    page = page.click("Nächste Abstimmung")
    assert "<td>101.1</td>" in page

    # 101.2
    page = page.click("Nächste Abstimmung")
    assert "<td>101.2</td>" in page

    # 102
    page = page.click("Nächste Abstimmung")
    assert "<td>102</td>" in page



from onegov.core.security import Private
from onegov.core.security import Public
from onegov.core.utils import normalize_for_url
from onegov.file.integration import render_depot_file
from onegov.form import Form
from onegov.swissvotes import _
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.forms import AttachmentsForm
from onegov.swissvotes.layouts import DeleteVoteLayout
from onegov.swissvotes.layouts import UploadVoteAttachemtsLayout
from onegov.swissvotes.layouts import VoteLayout
from onegov.swissvotes.layouts import VoteStrengthsLayout
from onegov.swissvotes.models import SwissVote
from onegov.swissvotes.models import SwissVoteFile


@SwissvotesApp.html(
    model=SwissVote,
    permission=Public,
    template='vote.pt'
)
def view_vote(self, request):
    query = request.session.query(SwissVote)
    prev = query.order_by(SwissVote.bfs_number.desc())
    prev = prev.filter(SwissVote.bfs_number < self.bfs_number).first()
    next = query.order_by(SwissVote.bfs_number.asc())
    next = next.filter(SwissVote.bfs_number > self.bfs_number).first()

    bfs_map = self.bfs_map(request.locale)
    bfs_map_host = self.bfs_map_host(request.locale)
    if bfs_map_host:
        request.content_security_policy.default_src |= {bfs_map_host}

    return {
        'layout': VoteLayout(self, request),
        'voting_text': self.get_file('voting_text'),
        'federal_council_message': self.get_file('federal_council_message'),
        'parliamentary_debate': self.get_file('parliamentary_debate'),
        'voting_booklet': self.get_file('voting_booklet'),
        'resolution': self.get_file('resolution'),
        'realization': self.get_file('realization'),
        'ad_analysis': self.get_file('ad_analysis'),
        'results_by_domain': self.get_file('results_by_domain'),
        'bfs_map': bfs_map,
        'prev': prev,
        'next': next,
    }


@SwissvotesApp.json(
    model=SwissVote,
    permission=Public,
    name='percentages'
)
def view_vote_percentages(self, request):
    return {
        'results': [
            {'text': request.translate(text), 'value': value}
            for text, value in self.percentages.items() if value is not None
        ],
        'title': self.title
    }


@SwissvotesApp.html(
    model=SwissVote,
    permission=Public,
    template='strengths.pt',
    name='strengths'
)
def view_vote_strengths(self, request):
    return {
        'layout': VoteStrengthsLayout(self, request)
    }


@SwissvotesApp.form(
    model=SwissVote,
    permission=Private,
    form=AttachmentsForm,
    template='form.pt',
    name='upload'
)
def upload_vote_attachments(self, request, form):
    layout = UploadVoteAttachemtsLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Attachments updated"), 'success')
        return request.redirect(request.link(self))

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'cancel': request.link(self)
    }


@SwissvotesApp.form(
    model=SwissVote,
    permission=Private,
    form=Form,
    template='form.pt',
    name='delete'
)
def delete_vote(self, request, form):
    layout = DeleteVoteLayout(self, request)

    if form.submitted(request):
        request.session.delete(self)
        request.message(_("Vote deleted"), 'success')
        return request.redirect(layout.votes_url)

    return {
        'layout': layout,
        'form': form,
        'subtitle': self.title,
        'message': _(
            'Do you really want to delete "${item}"?',
            mapping={'item': self.title}
        ),
        'button_text': _("Delete"),
        'button_class': 'alert',
        'cancel': request.link(self)
    }


@SwissvotesApp.view(
    model=SwissVoteFile,
    render=render_depot_file,
    permission=Public
)
def view_file(self, request):
    @request.after
    def set_filename(response):
        bfs_number = self.linked_swissvotes[0].bfs_number
        name = self.name.split('-')[0]
        extension = {'results_by_domain': 'xlsx'}.get(name, 'pdf')
        title = {
            'voting_text': _("Voting text"),
            'realization': _("Realization"),
            'federal_council_message': _("Federal council message"),
            'parliamentary_debate': _("Parliamentary debate"),
            'voting_booklet': _("Voting booklet"),
            'ad_analysis': _("Analysis of the advertising campaign"),
            'resolution': _("Resolution"),
            'results_by_domain': _(
                "Result by canton, district and municipality"
            )
        }.get(name, '')
        title = normalize_for_url(request.translate(title))
        response.headers['Content-Disposition'] = (
            f'inline; filename={bfs_number}-{title}.{extension}'
        )

    return self.reference.file



from io import BytesIO
from onegov.core.utils import module_path
from onegov.swissvotes.models import ColumnMapper
from onegov.swissvotes.models import SwissVote
from pytest import mark
from webtest import TestApp as Client
from webtest.forms import Upload
from xlsxwriter.workbook import Workbook


@mark.parametrize("file", [
    module_path('onegov.swissvotes', 'tests/fixtures/votes.xlsx'),
])
def test_update_votes(swissvotes_app, file):
    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    login = client.get('/auth/login')
    login.form['username'] = 'admin@example.org'
    login.form['password'] = 'hunter2'
    login.form.submit()

    with open(file, 'rb') as f:
        content = f.read()

    # Upload
    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Datensatz aktualisieren")
    manage.form['dataset'] = Upload(
        'votes.xlsx',
        content,
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    )
    manage = manage.form.submit().follow()
    assert "Datensatz aktualisiert (642 hinzugefügt, 0 geändert)" in manage

    session = swissvotes_app.session()
    vote = session.query(SwissVote).filter_by(bfs_number=82.2).one()
    assert str(vote.bfs_number) == '82.20'
    assert vote.date.isoformat() == '1920-03-21'
    assert vote.legislation_number == 25
    assert vote.decade.lower == 1920
    assert vote.decade.upper == 1929
    assert vote.title == (
        "Gegenentwurf zur Volksinitiative "
        "«für ein Verbot der Errichtung von Spielbanken»"
    )
    assert [str(pa) for pa in vote.policy_areas] == ['4.41.413', '4.44.443']
    assert str(vote.result_turnout) == '60.2323410000'
    assert vote.recommendations_parties['Nay'][0].name == 'sps'

    # Upload (unchanged)
    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Datensatz aktualisieren")
    manage.form['dataset'] = Upload(
        'votes.xlsx',
        content,
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    )
    manage = manage.form.submit().follow()
    assert "Datensatz aktualisiert (0 hinzugefügt, 0 geändert)" in manage

    # Download
    manage = client.get('/').maybe_follow().click("Abstimmungen")
    csv = manage.click("Datensatz herunterladen", index=0).body
    xlsx = manage.click("Datensatz herunterladen", index=1).body

    # Upload (roundtrip)
    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Datensatz aktualisieren")
    manage.form['dataset'] = Upload(
        'votes.xlsx',
        xlsx,
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    )
    manage = manage.form.submit().follow()
    assert "Datensatz aktualisiert (0 hinzugefügt, 0 geändert)" in manage

    assert csv == client.get('/votes/csv').body

    # Delete all votes
    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Alle Abstimmungen löschen")
    manage = manage.form.submit().follow()

    assert swissvotes_app.session().query(SwissVote).count() == 0


def test_update_votes_unknown_descriptors(swissvotes_app):
    client = Client(swissvotes_app)
    client.get('/locale/de_CH').follow()

    login = client.get('/auth/login')
    login.form['username'] = 'admin@example.org'
    login.form['password'] = 'hunter2'
    login.form.submit()

    file = BytesIO()
    workbook = Workbook(file)
    worksheet = workbook.add_worksheet()
    worksheet.write_row(0, 0, ColumnMapper().columns.values())
    worksheet.write_row(1, 0, [
        '100.1',  # anr
        '1.2.2008',  # datum
        '1',  # legislatur
        '2004-2008',  # legisjahr
        '2000-2009',  # jahrzent
        'titel',  # titel
        'stichwort',  # stichwort
        '2',  # anzahl
        '3',  # rechtsform
        '13',  # d1e1
        '',  # d1e2
        '',  # d1e3
        '12',  # d2e1
        '12.6',  # d2e2
        '',  # d2e3
        '12',  # d3e1
        '12.5',  # d3e2
        '12.55',  # d3e3
    ])
    workbook.close()
    file.seek(0)

    manage = client.get('/').maybe_follow().click("Abstimmungen")
    manage = manage.click("Datensatz aktualisieren")
    manage.form['dataset'] = Upload(
        'votes.xlsx',
        file.read(),
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    )
    manage = manage.form.submit().follow()
    assert "Datensatz aktualisiert (1 hinzugefügt, 0 geändert)" in manage
    assert "unbekannte Deskriptoren: 12.55, 12.6, 13" in manage



from dateutil.parser import parse
from decimal import Decimal
from onegov.form.fields import UploadField
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.swissvotes import _
from onegov.swissvotes.models import ColumnMapper
from onegov.swissvotes.models import SwissVote
from psycopg2.extras import NumericRange
from xlrd import open_workbook
from xlrd import XL_CELL_EMPTY
from xlrd import xldate


class SwissvoteDatasetField(UploadField):
    """ An upload field expecting a Swissvotes dataset (XLSX). """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('validators', [])
        kwargs['validators'].append(
            WhitelistedMimeType({
                'application/excel',
                'application/octet-stream',
                'application/vnd.ms-excel',
                'application/vnd.ms-office',
                'application/vnd.openxmlformats-officedocument.'
                'spreadsheetml.sheet',
                'application/zip'
            })
        )
        kwargs['validators'].append(FileSizeLimit(10 * 1024 * 1024))

        kwargs.setdefault('render_kw', {})
        kwargs['render_kw']['force_simple'] = True

        super().__init__(*args, **kwargs)

    def post_validate(self, form, validation_stopped):
        super(SwissvoteDatasetField, self).post_validate(
            form,
            validation_stopped
        )
        if validation_stopped:
            return

        errors = []
        data = []
        mapper = ColumnMapper()

        try:
            workbook = open_workbook(
                file_contents=self.raw_data[0].file.read()
            )
        except Exception:
            raise ValueError(_("Not a valid XLSX file."))

        if workbook.nsheets < 1:
            raise ValueError(_("No data."))

        sheet = workbook.sheet_by_index(0)

        if sheet.nrows <= 1:
            raise ValueError(_("No data."))

        headers = [column.value for column in sheet.row(0)]
        missing = set(mapper.columns.values()) - set(headers)
        if missing:
            raise ValueError(_(
                "Some columns are missing: ${columns}.",
                mapping={'columns': ', '.join(missing)}
            ))

        for index in range(1, sheet.nrows):
            row = sheet.row(index)
            vote = SwissVote()
            for (
                attribute, column, type_, nullable, precision, scale
            ) in mapper.items():
                cell = row[headers.index(column)]
                try:
                    if cell.ctype == XL_CELL_EMPTY:
                        value = None
                    elif type_ == 'TEXT':
                        value = str(cell.value)
                        value = '' if value == '.' else value
                    elif type_ == 'DATE':
                        if isinstance(cell.value, str):
                            value = parse(cell.value, dayfirst=True).date()
                        else:
                            value = xldate.xldate_as_datetime(
                                cell.value,
                                workbook.datemode
                            ).date()
                    elif type_ == 'INTEGER':
                        if isinstance(cell.value, str):
                            value = cell.value
                            value = '' if value == '.' else value
                            value = int(value) if value else None
                        else:
                            value = int(cell.value)
                    elif type_ == 'INT4RANGE':
                        value = NumericRange(*[
                            int(bound) for bound in cell.value.split('-')
                        ])
                    elif type_.startswith('NUMERIC'):
                        if isinstance(cell.value, str):
                            value = cell.value
                            value = '' if value == '.' else value
                            value = Decimal(str(value)) if value else None
                        else:
                            value = Decimal(str(cell.value))
                        if value is not None:
                            value = Decimal(
                                format(value, f'{precision}.{scale}f')
                            )

                except Exception:
                    errors.append((
                        index, column, f"'{value}' ≠ {type_.lower()}"
                    ))

                else:
                    if not nullable and value is None:
                        errors.append((index, column, "∅"))
                    mapper.set_value(vote, attribute, value)

            data.append(vote)

        if errors:
            raise ValueError(_(
                "Some cells contain invalid values: ${errors}.",
                mapping={
                    'errors': '; '.join([
                        '{}:{} {}'.format(*error) for error in errors
                    ])
                }
            ))

        self.data = data



from morepath import redirect
from onegov.core.crypto import random_token
from onegov.core.security import Private
from onegov.core.security import Public
from onegov.file.utils import as_fileintent
from onegov.form import Form
from onegov.swissvotes import _
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.swissvotes.forms import PageForm
from onegov.swissvotes.layouts import AddPageLayout
from onegov.swissvotes.layouts import DeletePageAttachmentLayout
from onegov.swissvotes.layouts import DeletePageLayout
from onegov.swissvotes.layouts import EditPageLayout
from onegov.swissvotes.layouts import PageAttachmentsLayout
from onegov.swissvotes.layouts import PageLayout
from onegov.swissvotes.models import TranslatablePage
from onegov.swissvotes.models import TranslatablePageFile
from onegov.swissvotes.models import TranslatablePageMove
from webob import exc


@SwissvotesApp.html(
    model=TranslatablePage,
    template='page.pt',
    permission=Public
)
def view_page(self, request):
    """ View a page. """

    files = [file for file in self.files if file.locale == request.locale]
    return {
        'layout': PageLayout(self, request),
        'files': files
    }


@SwissvotesApp.form(
    model=TranslatablePageCollection,
    permission=Private,
    form=PageForm,
    template='form.pt',
    name='add'
)
def add_page(self, request, form):
    request.include('quill')

    if form.submitted(request):
        page = TranslatablePage()
        form.update_model(page)
        request.session.add(page)
        request.message(_("Page added."), 'success')
        return request.redirect(request.link(page))

    return {
        'layout': AddPageLayout(self, request),
        'form': form
    }


@SwissvotesApp.form(
    model=TranslatablePage,
    permission=Private,
    form=PageForm,
    template='form.pt',
    name='edit'
)
def edit_page(self, request, form):
    request.include('quill')

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Page modified."), 'success')
        return request.redirect(request.link(self))

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': EditPageLayout(self, request),
        'form': form
    }


@SwissvotesApp.form(
    model=TranslatablePage,
    permission=Private,
    form=Form,
    template='form.pt',
    name='delete'
)
def delete_vote(self, request, form):
    layout = DeletePageLayout(self, request)

    if form.submitted(request):
        request.session.delete(self)
        request.message(_("Page deleted"), 'success')
        return request.redirect(layout.homepage_url)

    return {
        'layout': layout,
        'form': form,
        'subtitle': self.title,
        'message': _(
            'Do you really want to delete "${item}"?',
            mapping={'item': self.title}
        ),
        'button_text': _("Delete"),
        'button_class': 'alert',
        'cancel': request.link(self)
    }


@SwissvotesApp.view(
    model=TranslatablePageMove,
    permission=Private,
    request_method='PUT'
)
def move_page(self, request):
    request.assert_valid_csrf_token()
    self.execute()


@SwissvotesApp.html(
    model=TranslatablePage,
    template='attachments.pt',
    name='attachments',
    permission=Private
)
def view_page_attachments(self, request):
    """ View all attachments of a page and allow to drop new attachments. """

    layout = PageAttachmentsLayout(self, request)
    upload_url = layout.csrf_protected_url(request.link(self, name='upload'))
    files = [file for file in self.files if file.locale == request.locale]

    return {
        'layout': layout,
        'title': self.title,
        'subtitle': _("Attachments"),
        'upload_url': upload_url,
        'files': files,
        'notice': self,
    }


@SwissvotesApp.view(
    model=TranslatablePage,
    name='upload',
    permission=Private,
    request_method='POST'
)
def upload_page_attachment(self, request):
    """ Upload an attachment and add it to the page.

    Raises a HTTP 415 (Unsupported Media Type) if the file format is not
    supported.

    """

    request.assert_valid_csrf_token()

    attachment = TranslatablePageFile(id=random_token())
    attachment.name = '{}-{}'.format(
        request.locale,
        request.params['file'].filename
    )
    attachment.reference = as_fileintent(
        request.params['file'].file,
        request.params['file'].filename
    )

    if attachment.reference.content_type != 'application/pdf':
        raise exc.HTTPUnsupportedMediaType()

    self.files.append(attachment)
    request.message(_("Attachment added."), 'success')
    return redirect(request.link(self, 'attachments'))


@SwissvotesApp.form(
    model=TranslatablePageFile,
    name='delete',
    template='form.pt',
    permission=Private,
    form=Form
)
def delete_page_attachment(self, request, form):
    """ Delete an attachment. """

    layout = DeletePageAttachmentLayout(self, request)

    if form.submitted(request):
        url = request.link(self.linked_swissvotes_page[0], 'attachments')
        request.session.delete(self)
        request.message(_("Attachment deleted."), 'success')
        return redirect(url)

    return {
        'message': _(
            'Do you really want to delete "${item}"?',
            mapping={'item': self.filename}
        ),
        'layout': layout,
        'form': form,
        'title': self.filename,
        'subtitle': _("Delete"),
        'button_text': _("Delete"),
        'button_class': 'alert',
        'cancel': request.link(self)
    }



from cached_property import cached_property
from collections import OrderedDict
from onegov.swissvotes.models.vote import SwissVote


class ColumnMapper(object):
    """ Defines the columns used in the dataset and provides helper functions.

    """

    @cached_property
    def columns(self):
        """ The SwissVote attribute name and its column in the dataset.

        Attribute names starting with an ``!`` are used to indicate JSON
        attributes.
        """
        return OrderedDict((
            ('bfs_number', 'anr'),
            ('date', 'datum'),
            ('legislation_number', 'legislatur'),
            ('legislation_decade', 'legisjahr'),
            ('decade', 'jahrzehnt'),
            ('title', 'titel'),
            ('keyword', 'stichwort'),
            ('votes_on_same_day', 'anzahl'),
            ('_legal_form', 'rechtsform'),
            ('descriptor_1_level_1', 'd1e1'),
            ('descriptor_1_level_2', 'd1e2'),
            ('descriptor_1_level_3', 'd1e3'),
            ('descriptor_2_level_1', 'd2e1'),
            ('descriptor_2_level_2', 'd2e2'),
            ('descriptor_2_level_3', 'd2e3'),
            ('descriptor_3_level_1', 'd3e1'),
            ('descriptor_3_level_2', 'd3e2'),
            ('descriptor_3_level_3', 'd3e3'),
            ('_result_people_accepted', 'volk'),
            ('_result_cantons_accepted', 'stand'),
            ('_result', 'annahme'),
            ('result_eligible_voters', 'berecht'),
            ('result_votes_total', 'stimmen'),
            ('result_turnout', 'bet'),
            ('result_votes_empty', 'leer'),
            ('result_votes_invalid', 'ungultig'),
            ('result_votes_valid', 'gultig'),
            ('result_people_yeas', 'volkja'),
            ('result_people_nays', 'volknein'),
            ('result_people_yeas_p', 'volkja-proz'),
            ('result_cantons_yeas', 'kt-ja'),
            ('result_cantons_nays', 'kt-nein'),
            ('result_cantons_yeas_p', 'ktjaproz'),
            ('result_zh_eligible_voters', 'zh-berecht'),
            ('result_zh_votes_total', 'zh-stimmen'),
            ('result_zh_turnout', 'zh-bet'),
            ('result_zh_votes_valid', 'zh-gultig'),
            ('result_zh_yeas', 'zh-ja'),
            ('result_zh_nays', 'zh-nein'),
            ('result_zh_yeas_p', 'zh-japroz'),
            ('_result_zh_accepted', 'zh-annahme'),
            ('result_be_eligible_voters', 'be-berecht'),
            ('result_be_votes_total', 'be-stimmen'),
            ('result_be_turnout', 'be-bet'),
            ('result_be_votes_valid', 'be-gultig'),
            ('result_be_yeas', 'be-ja'),
            ('result_be_nays', 'be-nein'),
            ('result_be_yeas_p', 'be-japroz'),
            ('_result_be_accepted', 'be-annahme'),
            ('result_lu_eligible_voters', 'lu-berecht'),
            ('result_lu_votes_total', 'lu-stimmen'),
            ('result_lu_turnout', 'lu-bet'),
            ('result_lu_votes_valid', 'lu-gultig'),
            ('result_lu_yeas', 'lu-ja'),
            ('result_lu_nays', 'lu-nein'),
            ('result_lu_yeas_p', 'lu-japroz'),
            ('_result_lu_accepted', 'lu-annahme'),
            ('result_ur_eligible_voters', 'ur-berecht'),
            ('result_ur_votes_total', 'ur-stimmen'),
            ('result_ur_turnout', 'ur-bet'),
            ('result_ur_votes_valid', 'ur-gultig'),
            ('result_ur_yeas', 'ur-ja'),
            ('result_ur_nays', 'ur-nein'),
            ('result_ur_yeas_p', 'ur-japroz'),
            ('_result_ur_accepted', 'ur-annahme'),
            ('result_sz_eligible_voters', 'sz-berecht'),
            ('result_sz_votes_total', 'sz-stimmen'),
            ('result_sz_turnout', 'sz-bet'),
            ('result_sz_votes_valid', 'sz-gultig'),
            ('result_sz_yeas', 'sz-ja'),
            ('result_sz_nays', 'sz-nein'),
            ('result_sz_yeas_p', 'sz-japroz'),
            ('_result_sz_accepted', 'sz-annahme'),
            ('result_ow_eligible_voters', 'ow-berecht'),
            ('result_ow_votes_total', 'ow-stimmen'),
            ('result_ow_turnout', 'ow-bet'),
            ('result_ow_votes_valid', 'ow-gultig'),
            ('result_ow_yeas', 'ow-ja'),
            ('result_ow_nays', 'ow-nein'),
            ('result_ow_yeas_p', 'ow-japroz'),
            ('_result_ow_accepted', 'ow-annahme'),
            ('result_nw_eligible_voters', 'nw-berecht'),
            ('result_nw_votes_total', 'nw-stimmen'),
            ('result_nw_turnout', 'nw-bet'),
            ('result_nw_votes_valid', 'nw-gultig'),
            ('result_nw_yeas', 'nw-ja'),
            ('result_nw_nays', 'nw-nein'),
            ('result_nw_yeas_p', 'nw-japroz'),
            ('_result_nw_accepted', 'nw-annahme'),
            ('result_gl_eligible_voters', 'gl-berecht'),
            ('result_gl_votes_total', 'gl-stimmen'),
            ('result_gl_turnout', 'gl-bet'),
            ('result_gl_votes_valid', 'gl-gultig'),
            ('result_gl_yeas', 'gl-ja'),
            ('result_gl_nays', 'gl-nein'),
            ('result_gl_yeas_p', 'gl-japroz'),
            ('_result_gl_accepted', 'gl-annahme'),
            ('result_zg_eligible_voters', 'zg-berecht'),
            ('result_zg_votes_total', 'zg-stimmen'),
            ('result_zg_turnout', 'zg-bet'),
            ('result_zg_votes_valid', 'zg-gultig'),
            ('result_zg_yeas', 'zg-ja'),
            ('result_zg_nays', 'zg-nein'),
            ('result_zg_yeas_p', 'zg-japroz'),
            ('_result_zg_accepted', 'zg-annahme'),
            ('result_fr_eligible_voters', 'fr-berecht'),
            ('result_fr_votes_total', 'fr-stimmen'),
            ('result_fr_turnout', 'fr-bet'),
            ('result_fr_votes_valid', 'fr-gultig'),
            ('result_fr_yeas', 'fr-ja'),
            ('result_fr_nays', 'fr-nein'),
            ('result_fr_yeas_p', 'fr-japroz'),
            ('_result_fr_accepted', 'fr-annahme'),
            ('result_so_eligible_voters', 'so-berecht'),
            ('result_so_votes_total', 'so-stimmen'),
            ('result_so_turnout', 'so-bet'),
            ('result_so_votes_valid', 'so-gultig'),
            ('result_so_yeas', 'so-ja'),
            ('result_so_nays', 'so-nein'),
            ('result_so_yeas_p', 'so-japroz'),
            ('_result_so_accepted', 'so-annahme'),
            ('result_bs_eligible_voters', 'bs-berecht'),
            ('result_bs_votes_total', 'bs-stimmen'),
            ('result_bs_turnout', 'bs-bet'),
            ('result_bs_votes_valid', 'bs-gultig'),
            ('result_bs_yeas', 'bs-ja'),
            ('result_bs_nays', 'bs-nein'),
            ('result_bs_yeas_p', 'bs-japroz'),
            ('_result_bs_accepted', 'bs-annahme'),
            ('result_bl_eligible_voters', 'bl-berecht'),
            ('result_bl_votes_total', 'bl-stimmen'),
            ('result_bl_turnout', 'bl-bet'),
            ('result_bl_votes_valid', 'bl-gultig'),
            ('result_bl_yeas', 'bl-ja'),
            ('result_bl_nays', 'bl-nein'),
            ('result_bl_yeas_p', 'bl-japroz'),
            ('_result_bl_accepted', 'bl-annahme'),
            ('result_sh_eligible_voters', 'sh-berecht'),
            ('result_sh_votes_total', 'sh-stimmen'),
            ('result_sh_turnout', 'sh-bet'),
            ('result_sh_votes_valid', 'sh-gultig'),
            ('result_sh_yeas', 'sh-ja'),
            ('result_sh_nays', 'sh-nein'),
            ('result_sh_yeas_p', 'sh-japroz'),
            ('_result_sh_accepted', 'sh-annahme'),
            ('result_ar_eligible_voters', 'ar-berecht'),
            ('result_ar_votes_total', 'ar-stimmen'),
            ('result_ar_turnout', 'ar-bet'),
            ('result_ar_votes_valid', 'ar-gultig'),
            ('result_ar_yeas', 'ar-ja'),
            ('result_ar_nays', 'ar-nein'),
            ('result_ar_yeas_p', 'ar-japroz'),
            ('_result_ar_accepted', 'ar-annahme'),
            ('result_ai_eligible_voters', 'ai-berecht'),
            ('result_ai_votes_total', 'ai-stimmen'),
            ('result_ai_turnout', 'ai-bet'),
            ('result_ai_votes_valid', 'ai-gultig'),
            ('result_ai_yeas', 'ai-ja'),
            ('result_ai_nays', 'ai-nein'),
            ('result_ai_yeas_p', 'ai-japroz'),
            ('_result_ai_accepted', 'ai-annahme'),
            ('result_sg_eligible_voters', 'sg-berecht'),
            ('result_sg_votes_total', 'sg-stimmen'),
            ('result_sg_turnout', 'sg-bet'),
            ('result_sg_votes_valid', 'sg-gultig'),
            ('result_sg_yeas', 'sg-ja'),
            ('result_sg_nays', 'sg-nein'),
            ('result_sg_yeas_p', 'sg-japroz'),
            ('_result_sg_accepted', 'sg-annahme'),
            ('result_gr_eligible_voters', 'gr-berecht'),
            ('result_gr_votes_total', 'gr-stimmen'),
            ('result_gr_turnout', 'gr-bet'),
            ('result_gr_votes_valid', 'gr-gultig'),
            ('result_gr_yeas', 'gr-ja'),
            ('result_gr_nays', 'gr-nein'),
            ('result_gr_yeas_p', 'gr-japroz'),
            ('_result_gr_accepted', 'gr-annahme'),
            ('result_ag_eligible_voters', 'ag-berecht'),
            ('result_ag_votes_total', 'ag-stimmen'),
            ('result_ag_turnout', 'ag-bet'),
            ('result_ag_votes_valid', 'ag-gultig'),
            ('result_ag_yeas', 'ag-ja'),
            ('result_ag_nays', 'ag-nein'),
            ('result_ag_yeas_p', 'ag-japroz'),
            ('_result_ag_accepted', 'ag-annahme'),
            ('result_tg_eligible_voters', 'tg-berecht'),
            ('result_tg_votes_total', 'tg-stimmen'),
            ('result_tg_turnout', 'tg-bet'),
            ('result_tg_votes_valid', 'tg-gultig'),
            ('result_tg_yeas', 'tg-ja'),
            ('result_tg_nays', 'tg-nein'),
            ('result_tg_yeas_p', 'tg-japroz'),
            ('_result_tg_accepted', 'tg-annahme'),
            ('result_ti_eligible_voters', 'ti-berecht'),
            ('result_ti_votes_total', 'ti-stimmen'),
            ('result_ti_turnout', 'ti-bet'),
            ('result_ti_votes_valid', 'ti-gultig'),
            ('result_ti_yeas', 'ti-ja'),
            ('result_ti_nays', 'ti-nein'),
            ('result_ti_yeas_p', 'ti-japroz'),
            ('_result_ti_accepted', 'ti-annahme'),
            ('result_vd_eligible_voters', 'vd-berecht'),
            ('result_vd_votes_total', 'vd-stimmen'),
            ('result_vd_turnout', 'vd-bet'),
            ('result_vd_votes_valid', 'vd-gultig'),
            ('result_vd_yeas', 'vd-ja'),
            ('result_vd_nays', 'vd-nein'),
            ('result_vd_yeas_p', 'vd-japroz'),
            ('_result_vd_accepted', 'vd-annahme'),
            ('result_vs_eligible_voters', 'vs-berecht'),
            ('result_vs_votes_total', 'vs-stimmen'),
            ('result_vs_turnout', 'vs-bet'),
            ('result_vs_votes_valid', 'vs-gultig'),
            ('result_vs_yeas', 'vs-ja'),
            ('result_vs_nays', 'vs-nein'),
            ('result_vs_yeas_p', 'vs-japroz'),
            ('_result_vs_accepted', 'vs-annahme'),
            ('result_ne_eligible_voters', 'ne-berecht'),
            ('result_ne_votes_total', 'ne-stimmen'),
            ('result_ne_turnout', 'ne-bet'),
            ('result_ne_votes_valid', 'ne-gultig'),
            ('result_ne_yeas', 'ne-ja'),
            ('result_ne_nays', 'ne-nein'),
            ('result_ne_yeas_p', 'ne-japroz'),
            ('_result_ne_accepted', 'ne-annahme'),
            ('result_ge_eligible_voters', 'ge-berecht'),
            ('result_ge_votes_total', 'ge-stimmen'),
            ('result_ge_turnout', 'ge-bet'),
            ('result_ge_votes_valid', 'ge-gultig'),
            ('result_ge_yeas', 'ge-ja'),
            ('result_ge_nays', 'ge-nein'),
            ('result_ge_yeas_p', 'ge-japroz'),
            ('_result_ge_accepted', 'ge-annahme'),
            ('result_ju_eligible_voters', 'ju-berecht'),
            ('result_ju_votes_total', 'ju-stimmen'),
            ('result_ju_turnout', 'ju-bet'),
            ('result_ju_votes_valid', 'ju-gultig'),
            ('result_ju_yeas', 'ju-ja'),
            ('result_ju_nays', 'ju-nein'),
            ('result_ju_yeas_p', 'ju-japroz'),
            ('_result_ju_accepted', 'ju-annahme'),
            ('_department_in_charge', 'dep'),
            ('procedure_number', 'gesch_nr'),
            ('_position_federal_council', 'br-pos'),
            ('_position_parliament', 'bv-pos'),
            ('_position_national_council', 'nr-pos'),
            ('position_national_council_yeas', 'nrja'),
            ('position_national_council_nays', 'nrnein'),
            ('_position_council_of_states', 'sr-pos'),
            ('position_council_of_states_yeas', 'srja'),
            ('position_council_of_states_nays', 'srnein'),
            ('duration_federal_assembly', 'dauer_bv'),
            ('duration_post_federal_assembly', 'dauer_abst'),
            ('duration_initative_collection', 'i-dauer_samm'),
            ('duration_initative_federal_council', 'i-dauer_br'),
            ('duration_initative_total', 'i-dauer_tot'),
            ('duration_referendum_collection', 'fr-dauer_samm'),
            ('duration_referendum_total', 'fr-dauer_tot'),
            ('signatures_valid', 'unter_g'),
            ('signatures_invalid', 'unter_u'),
            ('!recommendations!fdp', 'p-fdp'),
            ('!recommendations!cvp', 'p-cvp'),
            ('!recommendations!sps', 'p-sps'),
            ('!recommendations!svp', 'p-svp'),
            ('!recommendations!evp', 'p-evp'),
            ('!recommendations!csp', 'p-ucsp'),
            ('!recommendations!pda', 'p-pda'),
            ('!recommendations!gps', 'p-gps'),
            ('!recommendations!sd', 'p-sd'),
            ('!recommendations!edu', 'p-edu'),
            ('!recommendations!fps', 'p-fps'),
            ('!recommendations!lega', 'p-lega'),
            ('!recommendations!kvp', 'p-kvp'),
            ('!recommendations!glp', 'p-glp'),
            ('!recommendations!bdp', 'p-bdp'),
            ('!recommendations!mcg', 'p-mcg'),
            ('!recommendations!lps', 'p-lps'),
            ('!recommendations!ldu', 'p-ldu'),
            ('!recommendations!poch', 'p-poch'),
            ('!recommendations!rep', 'p-rep'),
            ('!recommendations!sav', 'p-sav'),
            ('!recommendations!eco', 'p-eco'),
            ('!recommendations!sgv', 'p-sgv'),
            ('!recommendations!sbv-usp', 'p-sbv'),
            ('!recommendations!sgb', 'p-sgb'),
            ('!recommendations!travs', 'p-travs'),
            ('!recommendations!vsa', 'p-vsa'),
            ('!recommendations!vpod', 'p-vpod'),
            ('!recommendations!voev', 'p-voev'),
            ('!recommendations!tcs', 'p-tcs'),
            ('!recommendations!vcs', 'p-vcs'),
            ('!recommendations!acs', 'p-acs'),
            ('!recommendations!sbk', 'p-sbk'),
            ('!recommendations!ssv', 'p-ssv'),
            ('!recommendations!gem', 'p-gem'),
            ('!recommendations!kdk', 'p-kdk'),
            ('!recommendations!kkjpd', 'p-kkjpd'),
            ('!recommendations!gdk', 'p-gdk'),
            ('!recommendations!ldk', 'p-ldk'),
            ('!recommendations!vdk', 'p-vdk'),
            ('!recommendations!sodk', 'p-sodk'),
            ('!recommendations!endk', 'p-endk'),
            ('!recommendations!fdk', 'p-fdk'),
            ('!recommendations!edk', 'p-edk'),
            ('!recommendations!bpuk', 'p-bpuk'),
            ('recommendations_other_yes', 'p-others_yes'),
            ('recommendations_other_no', 'p-others_no'),
            ('recommendations_other_free', 'p-others_free'),
            ('!recommendations_divergent!bdp_ag', 'pdev-bdp_AG'),
            ('!recommendations_divergent!bdp_ai', 'pdev-bdp_AI'),
            ('!recommendations_divergent!bdp_ar', 'pdev-bdp_AR'),
            ('!recommendations_divergent!bdp_be', 'pdev-bdp_BE'),
            ('!recommendations_divergent!bdp_bl', 'pdev-bdp_BL'),
            ('!recommendations_divergent!bdp_bs', 'pdev-bdp_BS'),
            ('!recommendations_divergent!bdp_fr', 'pdev-bdp_FR'),
            ('!recommendations_divergent!bdp_ge', 'pdev-bdp_GE'),
            ('!recommendations_divergent!bdp_gl', 'pdev-bdp_GL'),
            ('!recommendations_divergent!bdp_gr', 'pdev-bdp_GR'),
            ('!recommendations_divergent!bdp_ju', 'pdev-bdp_JU'),
            ('!recommendations_divergent!bdp_lu', 'pdev-bdp_LU'),
            ('!recommendations_divergent!bdp_ne', 'pdev-bdp_NE'),
            ('!recommendations_divergent!bdp_nw', 'pdev-bdp_NW'),
            ('!recommendations_divergent!bdp_ow', 'pdev-bdp_OW'),
            ('!recommendations_divergent!bdp_sg', 'pdev-bdp_SG'),
            ('!recommendations_divergent!bdp_sh', 'pdev-bdp_SH'),
            ('!recommendations_divergent!bdp_so', 'pdev-bdp_SO'),
            ('!recommendations_divergent!bdp_sz', 'pdev-bdp_SZ'),
            ('!recommendations_divergent!bdp_tg', 'pdev-bdp_TG'),
            ('!recommendations_divergent!bdp_ti', 'pdev-bdp_TI'),
            ('!recommendations_divergent!bdp_ur', 'pdev-bdp_UR'),
            ('!recommendations_divergent!bdp_vd', 'pdev-bdp_VD'),
            ('!recommendations_divergent!bdp_vs', 'pdev-bdp_VS'),
            ('!recommendations_divergent!bdp_vsr', 'pdev-bdp_VSr'),
            ('!recommendations_divergent!bdp_vso', 'pdev-bdp_VSo'),
            ('!recommendations_divergent!bdp_zg', 'pdev-bdp_ZG'),
            ('!recommendations_divergent!bdp_zh', 'pdev-bdp_ZH'),
            ('!recommendations_divergent!jbdp_ch', 'pdev-jbdp_CH'),
            ('!recommendations_divergent!csp_fr', 'pdev-csp_FR'),
            ('!recommendations_divergent!csp_gr', 'pdev-csp_GR'),
            ('!recommendations_divergent!csp_ju', 'pdev-csp_JU'),
            ('!recommendations_divergent!csp_lu', 'pdev-csp_LU'),
            ('!recommendations_divergent!csp_ow', 'pdev-csp_OW'),
            ('!recommendations_divergent!csp_sg', 'pdev-csp_SG'),
            ('!recommendations_divergent!csp_vs', 'pdev-csp_VS'),
            ('!recommendations_divergent!csp_vsr', 'pdev-csp_VSr'),
            ('!recommendations_divergent!csp_vso', 'pdev-csp_VSo'),
            ('!recommendations_divergent!csp_zh', 'pdev-csp_ZH'),
            ('!recommendations_divergent!cvp-fr_ch', 'pdev-cvp_frauen'),
            ('!recommendations_divergent!cvp_ag', 'pdev-cvp_AG'),
            ('!recommendations_divergent!cvp_ai', 'pdev-cvp_AI'),
            ('!recommendations_divergent!cvp_ar', 'pdev-cvp_AR'),
            ('!recommendations_divergent!cvp_be', 'pdev-cvp_BE'),
            ('!recommendations_divergent!cvp_bl', 'pdev-cvp_BL'),
            ('!recommendations_divergent!cvp_bs', 'pdev-cvp_BS'),
            ('!recommendations_divergent!cvp_fr', 'pdev-cvp_FR'),
            ('!recommendations_divergent!cvp_ge', 'pdev-cvp_GE'),
            ('!recommendations_divergent!cvp_gl', 'pdev-cvp_GL'),
            ('!recommendations_divergent!cvp_gr', 'pdev-cvp_GR'),
            ('!recommendations_divergent!cvp_ju', 'pdev-cvp_JU'),
            ('!recommendations_divergent!cvp_lu', 'pdev-cvp_LU'),
            ('!recommendations_divergent!cvp_ne', 'pdev-cvp_NE'),
            ('!recommendations_divergent!cvp_nw', 'pdev-cvp_NW'),
            ('!recommendations_divergent!cvp_ow', 'pdev-cvp_OW'),
            ('!recommendations_divergent!cvp_sg', 'pdev-cvp_SG'),
            ('!recommendations_divergent!cvp_sh', 'pdev-cvp_SH'),
            ('!recommendations_divergent!cvp_so', 'pdev-cvp_SO'),
            ('!recommendations_divergent!cvp_sz', 'pdev-cvp_SZ'),
            ('!recommendations_divergent!cvp_tg', 'pdev-cvp_TG'),
            ('!recommendations_divergent!cvp_ti', 'pdev-cvp_TI'),
            ('!recommendations_divergent!cvp_ur', 'pdev-cvp_UR'),
            ('!recommendations_divergent!cvp_vd', 'pdev-cvp_VD'),
            ('!recommendations_divergent!cvp_vs', 'pdev-cvp_VS'),
            ('!recommendations_divergent!cvp_vsr', 'pdev-cvp_VSr'),
            ('!recommendations_divergent!cvp_vso', 'pdev-cvp_VSo'),
            ('!recommendations_divergent!cvp_zg', 'pdev-cvp_ZG'),
            ('!recommendations_divergent!cvp_zh', 'pdev-cvp_ZH'),
            ('!recommendations_divergent!jcvp_ch', 'pdev-jcvp_CH'),
            ('!recommendations_divergent!jcvp_ag', 'pdev-jcvp_AG'),
            ('!recommendations_divergent!jcvp_be', 'pdev-jcvp_BE'),
            ('!recommendations_divergent!jcvp_gr', 'pdev-jcvp_GR'),
            ('!recommendations_divergent!jcvp_lu', 'pdev-jcvp_LU'),
            ('!recommendations_divergent!jcvp_so', 'pdev-jcvp_SO'),
            ('!recommendations_divergent!jcvp_zh', 'pdev-jcvp_ZH'),
            ('!recommendations_divergent!edu_ag', 'pdev-edu_AG'),
            ('!recommendations_divergent!edu_ai', 'pdev-edu_AI'),
            ('!recommendations_divergent!edu_ar', 'pdev-edu_AR'),
            ('!recommendations_divergent!edu_be', 'pdev-edu_BE'),
            ('!recommendations_divergent!edu_bl', 'pdev-edu_BL'),
            ('!recommendations_divergent!edu_bs', 'pdev-edu_BS'),
            ('!recommendations_divergent!edu_fr', 'pdev-edu_FR'),
            ('!recommendations_divergent!edu_ge', 'pdev-edu_GE'),
            ('!recommendations_divergent!edu_gl', 'pdev-edu_GL'),
            ('!recommendations_divergent!edu_gr', 'pdev-edu_GR'),
            ('!recommendations_divergent!edu_ju', 'pdev-edu_JU'),
            ('!recommendations_divergent!edu_lu', 'pdev-edu_LU'),
            ('!recommendations_divergent!edu_ne', 'pdev-edu_NE'),
            ('!recommendations_divergent!edu_nw', 'pdev-edu_NW'),
            ('!recommendations_divergent!edu_ow', 'pdev-edu_OW'),
            ('!recommendations_divergent!edu_sg', 'pdev-edu_SG'),
            ('!recommendations_divergent!edu_sh', 'pdev-edu_SH'),
            ('!recommendations_divergent!edu_so', 'pdev-edu_SO'),
            ('!recommendations_divergent!edu_sz', 'pdev-edu_SZ'),
            ('!recommendations_divergent!edu_tg', 'pdev-edu_TG'),
            ('!recommendations_divergent!edu_ti', 'pdev-edu_TI'),
            ('!recommendations_divergent!edu_ur', 'pdev-edu_UR'),
            ('!recommendations_divergent!edu_vd', 'pdev-edu_VD'),
            ('!recommendations_divergent!edu_vs', 'pdev-edu_VS'),
            ('!recommendations_divergent!edu_vsr', 'pdev-edu_VSr'),
            ('!recommendations_divergent!edu_vso', 'pdev-edu_VSo'),
            ('!recommendations_divergent!edu_zg', 'pdev-edu_ZG'),
            ('!recommendations_divergent!edu_zh', 'pdev-edu_ZH'),
            ('!recommendations_divergent!evp_ag', 'pdev-evp_AG'),
            ('!recommendations_divergent!evp_ai', 'pdev-evp_AI'),
            ('!recommendations_divergent!evp_ar', 'pdev-evp_AR'),
            ('!recommendations_divergent!evp_be', 'pdev-evp_BE'),
            ('!recommendations_divergent!evp_bl', 'pdev-evp_BL'),
            ('!recommendations_divergent!evp_bs', 'pdev-evp_BS'),
            ('!recommendations_divergent!evp_fr', 'pdev-evp_FR'),
            ('!recommendations_divergent!evp_ge', 'pdev-evp_GE'),
            ('!recommendations_divergent!evp_gl', 'pdev-evp_GL'),
            ('!recommendations_divergent!evp_gr', 'pdev-evp_GR'),
            ('!recommendations_divergent!evp_ju', 'pdev-evp_JU'),
            ('!recommendations_divergent!evp_lu', 'pdev-evp_LU'),
            ('!recommendations_divergent!evp_ne', 'pdev-evp_NE'),
            ('!recommendations_divergent!evp_nw', 'pdev-evp_NW'),
            ('!recommendations_divergent!evp_ow', 'pdev-evp_OW'),
            ('!recommendations_divergent!evp_sg', 'pdev-evp_SG'),
            ('!recommendations_divergent!evp_sh', 'pdev-evp_SH'),
            ('!recommendations_divergent!evp_so', 'pdev-evp_SO'),
            ('!recommendations_divergent!evp_sz', 'pdev-evp_SZ'),
            ('!recommendations_divergent!evp_tg', 'pdev-evp_TG'),
            ('!recommendations_divergent!evp_ti', 'pdev-evp_TI'),
            ('!recommendations_divergent!evp_ur', 'pdev-evp_UR'),
            ('!recommendations_divergent!evp_vd', 'pdev-evp_VD'),
            ('!recommendations_divergent!evp_vs', 'pdev-evp_VS'),
            ('!recommendations_divergent!evp_zg', 'pdev-evp_ZG'),
            ('!recommendations_divergent!evp_zh', 'pdev-evp_ZH'),
            ('!recommendations_divergent!jevp_ch', 'pdev-jevp_CH'),
            ('!recommendations_divergent!fdp-fr_ch', 'pdev-fdp_Frauen'),
            ('!recommendations_divergent!fdp_ag', 'pdev-fdp_AG'),
            ('!recommendations_divergent!fdp_ai', 'pdev-fdp_AI'),
            ('!recommendations_divergent!fdp_ar', 'pdev-fdp_AR'),
            ('!recommendations_divergent!fdp_be', 'pdev-fdp_BE'),
            ('!recommendations_divergent!fdp_bl', 'pdev-fdp_BL'),
            ('!recommendations_divergent!fdp_bs', 'pdev-fdp_BS'),
            ('!recommendations_divergent!fdp_fr', 'pdev-fdp_FR'),
            ('!recommendations_divergent!fdp_ge', 'pdev-fdp_GE'),
            ('!recommendations_divergent!fdp_gl', 'pdev-fdp_GL'),
            ('!recommendations_divergent!fdp_gr', 'pdev-fdp_GR'),
            ('!recommendations_divergent!fdp_ju', 'pdev-fdp_JU'),
            ('!recommendations_divergent!fdp_lu', 'pdev-fdp_LU'),
            ('!recommendations_divergent!fdp_ne', 'pdev-fdp_NE'),
            ('!recommendations_divergent!fdp_nw', 'pdev-fdp_NW'),
            ('!recommendations_divergent!fdp_ow', 'pdev-fdp_OW'),
            ('!recommendations_divergent!fdp_sg', 'pdev-fdp_SG'),
            ('!recommendations_divergent!fdp_sh', 'pdev-fdp_SH'),
            ('!recommendations_divergent!fdp_so', 'pdev-fdp_SO'),
            ('!recommendations_divergent!fdp_sz', 'pdev-fdp_SZ'),
            ('!recommendations_divergent!fdp_tg', 'pdev-fdp_TG'),
            ('!recommendations_divergent!fdp_ti', 'pdev-fdp_TI'),
            ('!recommendations_divergent!fdp_ur', 'pdev-fdp_UR'),
            ('!recommendations_divergent!fdp_vd', 'pdev-fdp_VD'),
            ('!recommendations_divergent!fdp_vs', 'pdev-fdp_VS'),
            ('!recommendations_divergent!fdp_vsr', 'pdev-fdp_VSr'),
            ('!recommendations_divergent!fdp_vso', 'pdev-fdp_Vso'),
            ('!recommendations_divergent!fdp_zg', 'pdev-fdp_ZG'),
            ('!recommendations_divergent!fdp_zh', 'pdev-fdp_ZH'),
            ('!recommendations_divergent!jfdp_ch', 'pdev-jfdp_CH'),
            ('!recommendations_divergent!jfdp_ag', 'pdev-jfdp_AG'),
            ('!recommendations_divergent!jfdp_bl', 'pdev-jfdp_BL'),
            ('!recommendations_divergent!jfdp_fr', 'pdev-jfdp_FR'),
            ('!recommendations_divergent!jfdp_gr', 'pdev-jfdp_GR'),
            ('!recommendations_divergent!jfdp_ju', 'pdev-jfdp_JU'),
            ('!recommendations_divergent!jfdp_lu', 'pdev-jfdp_LU'),
            ('!recommendations_divergent!jfdp_sh', 'pdev-jfdp_SH'),
            ('!recommendations_divergent!jfdp_ti', 'pdev-jfdp_TI'),
            ('!recommendations_divergent!jfdp_vd', 'pdev-jfdp_VD'),
            ('!recommendations_divergent!jfdp_zh', 'pdev-jfdp_ZH'),
            ('!recommendations_divergent!fps_ag', 'pdev-fps_AG'),
            ('!recommendations_divergent!fps_ai', 'pdev-fps_AI'),
            ('!recommendations_divergent!fps_be', 'pdev-fps_BE'),
            ('!recommendations_divergent!fps_bl', 'pdev-fps_BL'),
            ('!recommendations_divergent!fps_bs', 'pdev-fps_BS'),
            ('!recommendations_divergent!fps_sg', 'pdev-fps_SG'),
            ('!recommendations_divergent!fps_sh', 'pdev-fps_SH'),
            ('!recommendations_divergent!fps_so', 'pdev-fps_SO'),
            ('!recommendations_divergent!fps_tg', 'pdev-fps_TG'),
            ('!recommendations_divergent!fps_zh', 'pdev-fps_ZH'),
            ('!recommendations_divergent!glp_ag', 'pdev-glp_AG'),
            ('!recommendations_divergent!glp_ai', 'pdev-glp_AI'),
            ('!recommendations_divergent!glp_ar', 'pdev-glp_AR'),
            ('!recommendations_divergent!glp_be', 'pdev-glp_BE'),
            ('!recommendations_divergent!glp_bl', 'pdev-glp_BL'),
            ('!recommendations_divergent!glp_bs', 'pdev-glp_BS'),
            ('!recommendations_divergent!glp_fr', 'pdev-glp_FR'),
            ('!recommendations_divergent!glp_ge', 'pdev-glp_GE'),
            ('!recommendations_divergent!glp_gl', 'pdev-glp_GL'),
            ('!recommendations_divergent!glp_gr', 'pdev-glp_GR'),
            ('!recommendations_divergent!glp_ju', 'pdev-glp_JU'),
            ('!recommendations_divergent!glp_lu', 'pdev-glp_LU'),
            ('!recommendations_divergent!glp_ne', 'pdev-glp_NE'),
            ('!recommendations_divergent!glp_nw', 'pdev-glp_NW'),
            ('!recommendations_divergent!glp_ow', 'pdev-glp_OW'),
            ('!recommendations_divergent!glp_sg', 'pdev-glp_SG'),
            ('!recommendations_divergent!glp_sh', 'pdev-glp_SH'),
            ('!recommendations_divergent!glp_so', 'pdev-glp_SO'),
            ('!recommendations_divergent!glp_sz', 'pdev-glp_SZ'),
            ('!recommendations_divergent!glp_tg', 'pdev-glp_TG'),
            ('!recommendations_divergent!glp_ti', 'pdev-glp_TI'),
            ('!recommendations_divergent!glp_ur', 'pdev-glp_UR'),
            ('!recommendations_divergent!glp_vd', 'pdev-glp_VD'),
            ('!recommendations_divergent!glp_vs', 'pdev-glp_VS'),
            ('!recommendations_divergent!glp_vsr', 'pdev-glp_VSr'),
            ('!recommendations_divergent!glp_vso', 'pdev-glp_VSo'),
            ('!recommendations_divergent!glp_zg', 'pdev-glp_ZG'),
            ('!recommendations_divergent!glp_zh', 'pdev-glp_ZH'),
            ('!recommendations_divergent!jglp_ch', 'pdev-jglp_CH'),
            ('!recommendations_divergent!gps_ag', 'pdev-gps_AG'),
            ('!recommendations_divergent!gps_ai', 'pdev-gps_AI'),
            ('!recommendations_divergent!gps_ar', 'pdev-gps_AR'),
            ('!recommendations_divergent!gps_be', 'pdev-gps_BE'),
            ('!recommendations_divergent!gps_bl', 'pdev-gps_BL'),
            ('!recommendations_divergent!gps_bs', 'pdev-gps_BS'),
            ('!recommendations_divergent!gps_fr', 'pdev-gps_FR'),
            ('!recommendations_divergent!gps_ge', 'pdev-gps_GE'),
            ('!recommendations_divergent!gps_gl', 'pdev-gps_GL'),
            ('!recommendations_divergent!gps_gr', 'pdev-gps_GR'),
            ('!recommendations_divergent!gps_ju', 'pdev-gps_JU'),
            ('!recommendations_divergent!gps_lu', 'pdev-gps_LU'),
            ('!recommendations_divergent!gps_ne', 'pdev-gps_NE'),
            ('!recommendations_divergent!gps_nw', 'pdev-gps_NW'),
            ('!recommendations_divergent!gps_ow', 'pdev-gps_OW'),
            ('!recommendations_divergent!gps_sg', 'pdev-gps_SG'),
            ('!recommendations_divergent!gps_sh', 'pdev-gps_SH'),
            ('!recommendations_divergent!gps_so', 'pdev-gps_SO'),
            ('!recommendations_divergent!gps_sz', 'pdev-gps_SZ'),
            ('!recommendations_divergent!gps_tg', 'pdev-gps_TG'),
            ('!recommendations_divergent!gps_ti', 'pdev-gps_TI'),
            ('!recommendations_divergent!gps_ur', 'pdev-gps_UR'),
            ('!recommendations_divergent!gps_vd', 'pdev-gps_VD'),
            ('!recommendations_divergent!gps_vs', 'pdev-gps_VS'),
            ('!recommendations_divergent!gps_vsr', 'pdev-gps_VSr'),
            ('!recommendations_divergent!gps_vso', 'pdev-gps_VSo'),
            ('!recommendations_divergent!gps_zg', 'pdev-gps_ZG'),
            ('!recommendations_divergent!gps_zh', 'pdev-gps_ZH'),
            ('!recommendations_divergent!jgps_ch', 'pdev-jgps_CH'),
            ('!recommendations_divergent!kvp_sg', 'pdev-kvp_SG'),
            ('!recommendations_divergent!lps_be', 'pdev-lps_BE'),
            ('!recommendations_divergent!lps_bl', 'pdev-lps_BL'),
            ('!recommendations_divergent!lps_bs', 'pdev-lps_BS'),
            ('!recommendations_divergent!lps_fr', 'pdev-lps_FR'),
            ('!recommendations_divergent!lps_ge', 'pdev-lps_GE'),
            ('!recommendations_divergent!lps_ju', 'pdev-lps_JU'),
            ('!recommendations_divergent!lps_ne', 'pdev-lps_NE'),
            ('!recommendations_divergent!lps_sg', 'pdev-lps_SG'),
            ('!recommendations_divergent!lps_ti', 'pdev-lps_TI'),
            ('!recommendations_divergent!lps_vd', 'pdev-lps_VD'),
            ('!recommendations_divergent!lps_vs', 'pdev-lps_VS'),
            ('!recommendations_divergent!lps_zh', 'pdev-lps_ZH'),
            ('!recommendations_divergent!jlps_ch', 'pdev-jlps_CH'),
            ('!recommendations_divergent!jlps_so', 'pdev-jlps_SO'),
            ('!recommendations_divergent!jlps_ti', 'pdev-jlps_TI'),
            ('!recommendations_divergent!ldu_ag', 'pdev-ldu_AG'),
            ('!recommendations_divergent!ldu_ar', 'pdev-ldu_AR'),
            ('!recommendations_divergent!ldu_be', 'pdev-ldu_BE'),
            ('!recommendations_divergent!ldu_bl', 'pdev-ldu_BL'),
            ('!recommendations_divergent!ldu_bs', 'pdev-ldu_BS'),
            ('!recommendations_divergent!ldu_gr', 'pdev-ldu_GR'),
            ('!recommendations_divergent!ldu_lu', 'pdev-ldu_LU'),
            ('!recommendations_divergent!ldu_ne', 'pdev-ldu_NE'),
            ('!recommendations_divergent!ldu_sg', 'pdev-ldu_SG'),
            ('!recommendations_divergent!ldu_sh', 'pdev-ldu_SH'),
            ('!recommendations_divergent!ldu_so', 'pdev-ldu_SO'),
            ('!recommendations_divergent!ldu_tg', 'pdev-ldu_TG'),
            ('!recommendations_divergent!ldu_vd', 'pdev-ldu_VD'),
            ('!recommendations_divergent!ldu_zg', 'pdev-ldu_ZG'),
            ('!recommendations_divergent!ldu_zh', 'pdev-ldu_ZH'),
            ('!recommendations_divergent!jldu_ch', 'pdev-jldu_CH'),
            ('!recommendations_divergent!poch_so', 'pdev-poch_SO'),
            ('!recommendations_divergent!poch_zh', 'pdev-poch_ZH'),
            ('!recommendations_divergent!pda_be', 'pdev-pda_BE'),
            ('!recommendations_divergent!pda_bl', 'pdev-pda_BL'),
            ('!recommendations_divergent!pda_bs', 'pdev-pda_BS'),
            ('!recommendations_divergent!pda_ge', 'pdev-pda_GE'),
            ('!recommendations_divergent!pda_ju', 'pdev-pda_JU'),
            ('!recommendations_divergent!pda_ne', 'pdev-pda_NE'),
            ('!recommendations_divergent!pda_sg', 'pdev-pda_SG'),
            ('!recommendations_divergent!pda_ti', 'pdev-pda_TI'),
            ('!recommendations_divergent!pda_vd', 'pdev-pda_VD'),
            ('!recommendations_divergent!pda_zh', 'pdev-pda_ZH'),
            ('!recommendations_divergent!jpda_ch', 'pdev-jpda_CH'),
            ('!recommendations_divergent!rep_ag', 'pdev-rep_AG'),
            ('!recommendations_divergent!rep_ge', 'pdev-rep_GE'),
            ('!recommendations_divergent!rep_ne', 'pdev-rep_NE'),
            ('!recommendations_divergent!rep_tg', 'pdev-rep_TG'),
            ('!recommendations_divergent!rep_vd', 'pdev-rep_VD'),
            ('!recommendations_divergent!rep_zh', 'pdev-rep_ZH'),
            ('!recommendations_divergent!sd_ag', 'pdev-sd_AG'),
            ('!recommendations_divergent!sd_be', 'pdev-sd_BE'),
            ('!recommendations_divergent!sd_bl', 'pdev-sd_BL'),
            ('!recommendations_divergent!sd_bs', 'pdev-sd_BS'),
            ('!recommendations_divergent!sd_fr', 'pdev-sd_FR'),
            ('!recommendations_divergent!sd_ge', 'pdev-sd_GE'),
            ('!recommendations_divergent!sd_gr', 'pdev-sd_GR'),
            ('!recommendations_divergent!sd_lu', 'pdev-sd_LU'),
            ('!recommendations_divergent!sd_ne', 'pdev-sd_NE'),
            ('!recommendations_divergent!sd_sg', 'pdev-sd_SG'),
            ('!recommendations_divergent!sd_so', 'pdev-sd_SO'),
            ('!recommendations_divergent!sd_tg', 'pdev-sd_TG'),
            ('!recommendations_divergent!sd_ti', 'pdev-sd_TI'),
            ('!recommendations_divergent!sd_vd', 'pdev-sd_VD'),
            ('!recommendations_divergent!sd_zh', 'pdev-sd_ZH'),
            ('!recommendations_divergent!jsd_ch', 'pdev-jsd_CH'),
            ('!recommendations_divergent!sps_ag', 'pdev-sps_AG'),
            ('!recommendations_divergent!sps_ai', 'pdev-sps_AI'),
            ('!recommendations_divergent!sps_ar', 'pdev-sps_AR'),
            ('!recommendations_divergent!sps_be', 'pdev-sps_BE'),
            ('!recommendations_divergent!sps_bl', 'pdev-sps_BL'),
            ('!recommendations_divergent!sps_bs', 'pdev-sps_BS'),
            ('!recommendations_divergent!sps_fr', 'pdev-sps_FR'),
            ('!recommendations_divergent!sps_ge', 'pdev-sps_GE'),
            ('!recommendations_divergent!sps_gl', 'pdev-sps_GL'),
            ('!recommendations_divergent!sps_gr', 'pdev-sps_GR'),
            ('!recommendations_divergent!sps_ju', 'pdev-sps_JU'),
            ('!recommendations_divergent!sps_lu', 'pdev-sps_LU'),
            ('!recommendations_divergent!sps_ne', 'pdev-sps_NE'),
            ('!recommendations_divergent!sps_nw', 'pdev-sps_NW'),
            ('!recommendations_divergent!sps_ow', 'pdev-sps_OW'),
            ('!recommendations_divergent!sps_sg', 'pdev-sps_SG'),
            ('!recommendations_divergent!sps_sh', 'pdev-sps_SH'),
            ('!recommendations_divergent!sps_so', 'pdev-sps_SO'),
            ('!recommendations_divergent!sps_sz', 'pdev-sps_SZ'),
            ('!recommendations_divergent!sps_tg', 'pdev-sps_TG'),
            ('!recommendations_divergent!sps_ti', 'pdev-sps_TI'),
            ('!recommendations_divergent!sps_ur', 'pdev-sps_UR'),
            ('!recommendations_divergent!sps_vd', 'pdev-sps_VD'),
            ('!recommendations_divergent!sps_vs', 'pdev-sps_VS'),
            ('!recommendations_divergent!sps_vsr', 'pdev-sps_VSr'),
            ('!recommendations_divergent!sps_vso', 'pdev-sps_VSo'),
            ('!recommendations_divergent!sps_zg', 'pdev-sps_ZG'),
            ('!recommendations_divergent!sps_zh', 'pdev-sps_ZH'),
            ('!recommendations_divergent!juso_ch', 'pdev-juso_CH'),
            ('!recommendations_divergent!juso_be', 'pdev-juso_BE'),
            ('!recommendations_divergent!juso_ge', 'pdev-juso_GE'),
            ('!recommendations_divergent!juso_ju', 'pdev-juso_JU'),
            ('!recommendations_divergent!juso_ti', 'pdev-juso_TI'),
            ('!recommendations_divergent!juso_vs', 'pdev-juso_VS'),
            ('!recommendations_divergent!juso_zh', 'pdev-juso_ZH'),
            ('!recommendations_divergent!svp_ag', 'pdev-svp_AG'),
            ('!recommendations_divergent!svp_ai', 'pdev-svp_AI'),
            ('!recommendations_divergent!svp_ar', 'pdev-svp_AR'),
            ('!recommendations_divergent!svp_be', 'pdev-svp_BE'),
            ('!recommendations_divergent!svp_bl', 'pdev-svp_BL'),
            ('!recommendations_divergent!svp_bs', 'pdev-svp_BS'),
            ('!recommendations_divergent!svp_fr', 'pdev-svp_FR'),
            ('!recommendations_divergent!svp_ge', 'pdev-svp_GE'),
            ('!recommendations_divergent!svp_gl', 'pdev-svp_GL'),
            ('!recommendations_divergent!svp_gr', 'pdev-svp_GR'),
            ('!recommendations_divergent!svp_ju', 'pdev-svp_JU'),
            ('!recommendations_divergent!svp_lu', 'pdev-svp_LU'),
            ('!recommendations_divergent!svp_ne', 'pdev-svp_NE'),
            ('!recommendations_divergent!svp_nw', 'pdev-svp_NW'),
            ('!recommendations_divergent!svp_ow', 'pdev-svp_OW'),
            ('!recommendations_divergent!svp_sg', 'pdev-svp_SG'),
            ('!recommendations_divergent!svp_sh', 'pdev-svp_SH'),
            ('!recommendations_divergent!svp_so', 'pdev-svp_SO'),
            ('!recommendations_divergent!svp_sz', 'pdev-svp_SZ'),
            ('!recommendations_divergent!svp_tg', 'pdev-svp_TG'),
            ('!recommendations_divergent!svp_ti', 'pdev-svp_TI'),
            ('!recommendations_divergent!svp_ur', 'pdev-svp_UR'),
            ('!recommendations_divergent!svp_vd', 'pdev-svp_VD'),
            ('!recommendations_divergent!svp_vs', 'pdev-svp_VS'),
            ('!recommendations_divergent!svp_vsr', 'pdev-svp_VSr'),
            ('!recommendations_divergent!svp_vso', 'pdev-svp_VSo'),
            ('!recommendations_divergent!svp_zg', 'pdev-svp_ZG'),
            ('!recommendations_divergent!svp_zh', 'pdev-svp_ZH'),
            ('!recommendations_divergent!jsvp_ch', 'pdev-jsvp_CH'),
            ('!recommendations_divergent!jsvp_ag', 'pdev-jsvp_AG'),
            ('!recommendations_divergent!jsvp_be', 'pdev-jsvp_BE'),
            ('!recommendations_divergent!jsvp_ge', 'pdev-jsvp_GE'),
            ('!recommendations_divergent!jsvp_sh', 'pdev-jsvp_SH'),
            ('!recommendations_divergent!jsvp_ur', 'pdev-jsvp_UR'),
            ('!recommendations_divergent!jsvp_zh', 'pdev-jsvp_ZH'),
            ('!recommendations_divergent!sgb_ag', 'pdev-sgb_AG'),
            ('!recommendations_divergent!sgb_ju', 'pdev-sgb_JU'),
            ('!recommendations_divergent!sgb_vs', 'pdev-sgb_VS'),
            ('!recommendations_divergent!sgv_ag', 'pdev-sgv_AG'),
            ('!recommendations_divergent!sgv_bs', 'pdev-sgv_BS'),
            ('!recommendations_divergent!sgv_sh', 'pdev-sgv_SH'),
            ('!recommendations_divergent!vpod_ge', 'pdev-vpod_GE'),
            ('!recommendations_divergent!vpod_vd', 'pdev-vpod_VD'),
            ('national_council_election_year', 'nr-wahl'),
            ('national_council_share_fdp', 'w-fdp'),
            ('national_council_share_cvp', 'w-cvp'),
            ('national_council_share_sp', 'w-sp'),
            ('national_council_share_svp', 'w-svp'),
            ('national_council_share_lps', 'w-lps'),
            ('national_council_share_ldu', 'w-ldu'),
            ('national_council_share_evp', 'w-evp'),
            ('national_council_share_csp', 'w-csp'),
            ('national_council_share_pda', 'w-pda'),
            ('national_council_share_poch', 'w-poch'),
            ('national_council_share_gps', 'w-gps'),
            ('national_council_share_sd', 'w-sd'),
            ('national_council_share_rep', 'w-rep'),
            ('national_council_share_edu', 'w-edu'),
            ('national_council_share_fps', 'w-fps'),
            ('national_council_share_lega', 'w-lega'),
            ('national_council_share_kvp', 'w-kvp'),
            ('national_council_share_glp', 'w-glp'),
            ('national_council_share_bdp', 'w-bdp'),
            ('national_council_share_mcg', 'w-mcg'),
            ('national_council_share_ubrige', 'w-ubrige'),
            ('national_council_share_yeas', 'ja-lager'),
            ('national_council_share_nays', 'nein-lager'),
            ('national_council_share_neutral', 'neutral'),
            ('national_council_share_vague', 'unbestimmt'),
            ('initiator', 'urheber'),
            ('anneepolitique', 'anneepolitique'),
            ('bfs_map_de', 'bfsmap-de'),
            ('bfs_map_fr', 'bfsmap-fr'),
        ))

    def set_value(self, vote, attribute, value):
        """ Set the given value of a vote. """
        if attribute.startswith('!'):
            unused, attribute, key = attribute.split('!')
            if getattr(vote, attribute) is None:
                setattr(vote, attribute, {})
            getattr(vote, attribute)[key] = value
        else:
            setattr(vote, attribute, value)

    def get_value(self, vote, attribute):
        """ Get the given value of a vote. """
        if attribute.startswith('!'):
            unused, attribute, key = attribute.split('!')
            return (getattr(vote, attribute) or {}).get(key)
        return getattr(vote, attribute)

    def get_values(self, vote):
        """ Get all values of a vote in order. """
        for attribute in self.columns.keys():
            yield self.get_value(vote, attribute)

    def get_items(self, vote):
        """ Get all names and values of a vote in order. """
        for attribute in self.columns.keys():
            yield attribute, self.get_value(vote, attribute)

    def items(self):
        """ Returns the attributes and column names together with additional
        information (type, nullable, precision, scale).

        """
        for attribute, column in self.columns.items():
            if attribute.startswith('!'):
                type_ = 'INTEGER'
                nullable = True
                precision = None
                scale = None
            else:
                name = attribute.lstrip('_')
                table_column = SwissVote.__table__.columns[name]
                type_ = str(table_column.type)
                nullable = table_column.nullable
                precision = getattr(table_column.type, 'precision', None)
                scale = getattr(table_column.type, 'scale', None)
            yield attribute, column, type_, nullable, precision, scale
